package atc.safefly.communication;

import java.io.BufferedWriter;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileWriter;
import java.io.IOException;
import java.io.InputStream;
import java.text.DateFormat;
import java.util.Date;
import java.util.List;
import java.util.Properties;
import java.util.Scanner;
import java.util.concurrent.LinkedBlockingQueue;

import org.apache.log4j.Logger;
import org.apache.log4j.PropertyConfigurator;

import atc.safefly.communication.common.SafeflyConnector;
import atc.safefly.communication.common.packet.ApplicationPacket;
import atc.safefly.communication.common.packet.Packet;
import atc.safefly.communication.common.packet.PacketConnectionOwner;
import atc.safefly.communication.common.packet.PacketInterface;
import atc.safefly.communication.common.packet.VisualizationPacket;
import atc.safefly.communication.common.serial.UniversalSerialConnector;
import atc.safefly.communication.utils.ApplicationData;
import atc.safefly.communication.utils.ZaonData;
import atc.safefly.ontology.Group;
import atc.safefly.ontology.Groups;
import atc.safefly.ontology.SerialPortSettings;
import atc.safefly.utils.ByteUtils;
import atc.safefly.utils.SafeflyConstants.Cmd_ID;
import atc.safefly.utils.SafeflyConstants.Cmd_Type;
import atc.safefly.utils.SafeflyTelemetry;
import atc.safefly.utils.XmlLoader;

/**
 * Class for processing communication with Integra Glass Cockpit - subscribes
 * for data and parses incoming data packets
 * 
 */
public class IntegraCommunicator {

	private static final Logger LOGGER = Logger.getLogger(IntegraCommunicator.class);
	

	// ///////////////////////////////////////////////
	// ///// COMMUNICATION ESTABLISHMENT
	// ///////////////////////////////////////////////

	/** Connector for communication with Integra GlassCockpit */
	protected SafeflyConnector integraConnector;
	
	/** Connector for visualization on Integra GlassCockpit */
	protected SafeflyConnector visualizationConnector;

	/** Name of the integra serial port to connect to */
	protected String integraSerialName;
	
	/** Name of visualization Integra serial port to fork to */
	protected String visualizationSerialName;
	
	/** baud rate of the integra serial communication */
	protected int integraBaudRate;

	/** Name of the modem serial port to connect to */
	protected String modemSerialName;
	
	/** baud rate of the modem serial communication */
	protected int modemBaudRate;

	protected SafeflyConnector dataConnector;

	/** owner for arrived data processing */
	protected IntegraCommOwner owner;

	// //////////////////////////////////////////
	// ///// PACKET SENDING
	// //////////////////////////////////////////

	protected LinkedBlockingQueue<PacketInterface> sendPacketQueue = new LinkedBlockingQueue<PacketInterface>();
	protected volatile boolean ackReceived = false;

	/** time [ms] to wait between sending two requests */
	protected static final int INTER_PACKET_WAIT_TIME = 1500;

	/** period [ms] for sending telemetry to ground station */
	protected static final int TELEMETRY_SEND_PERIOD = 2000;

	// //////////////////////////////////////////
	// //////// LOGGING
	// //////////////////////////////////////////

	/** Logging file */
	protected BufferedWriter logFile;
	private static final String LOG_DIR = "gumstix_client_log_files";
	private static final String LOG_FILE_PREFIX = "integra_log_";

	// ///////////////////////////////////////////
	// /////// COMMAND LINE DEFAULT ARGS
	// ///////////////////////////////////////////

	private static final String DEFAULT_INTEGRA_PORT_XML_LOCATION = "xml/serial_port.xml";
	private static final String DEFAULT_CHANNELS_XML_LOCATION = "xml/integra_channels.xml";
	private static final String DEFAULT_MODEM_PORT_XML_LOCATION = "xml/modem_port.xml";

	private Thread integraThread;
	private Thread modemThread;

	// ///////////////////////////////////////////
	// ////// DATA
	// ///////////////////////////////////////////
	
	protected Groups integraGroups;

	protected SafeflyTelemetry curState = new SafeflyTelemetry();
	protected ZaonData zaonData = new ZaonData();

	/**
	 * Constructor
	 * 
	 * @param serialName
	 *            - name of the Integra serial port
	 * @param baudRate
	 *            - used serial baud rate
	 */
	public IntegraCommunicator(SerialPortSettings integraCom, SerialPortSettings modemCom) {
		LOGGER.info("initialization...");
		this.integraSerialName = integraCom.getPort();
		this.integraBaudRate = integraCom.getBaudRate();
		if(modemCom != null){
			this.modemSerialName = modemCom.getPort();
			this.modemBaudRate = modemCom.getBaudRate();
		}
	}

	/**
	 * Constructor
	 * 
	 * @param integraCom
	 * 			- Settings for Integra serial port
	 * @param owner
	 * 			- Integra connector owner for processing of incoming packets
	 * @param groups
	 * 			/ Integra measure protocol {@link Groups} for registration 
	 */
	public IntegraCommunicator(SerialPortSettings integraCom, IntegraCommOwner owner, Groups groups) {
		LOGGER.info("initialization...");
		this.integraSerialName = integraCom.getPort();
		this.integraBaudRate = integraCom.getBaudRate();
		this.owner = owner;
		this.integraGroups = groups;
	}
	
	/**
	 * Constructor for combination of simulated measure protocol
	 * and real visualization on Integra
	 * 
	 * @param integraCom
	 * 			- Settings for Integra serial port
	 * @param visCom
	 * 			- Special settings for visualization packets
	 * @param owner
	 * 			- Integra connector owner for processing of incoming packets
	 * @param groups
	 * 			/ Integra measure protocol {@link Groups} for registration 
	 */
	public IntegraCommunicator(SerialPortSettings integraCom, SerialPortSettings visCom, IntegraCommOwner owner, Groups groups) {
		this(integraCom, owner, groups);
		visualizationSerialName = visCom.getPort();
		LOGGER.info("Special visualization port set: "+visualizationSerialName);
	}

	/** Establish connection */
	public boolean connect() {
		boolean isConnected = false;
		UniversalSerialConnector serialConn = new UniversalSerialConnector(integraSerialName, integraBaudRate);
		UniversalSerialConnector modemCon = new UniversalSerialConnector(modemSerialName, modemBaudRate);

		integraConnector = new SafeflyConnector(serialConn, new PacketConnectionOwner() {

			@Override
			public void incommingPacket(PacketInterface packet) {
				processPacket(packet);
			}
		});
		
		// Special visualization port for hybrid simulation with real integra visualization 
		if(visualizationSerialName != null) {
			visualizationConnector = new SafeflyConnector(new UniversalSerialConnector(visualizationSerialName, integraBaudRate), new PacketConnectionOwner() {
				
				@Override
				public void incommingPacket(PacketInterface packet) {
					processPacket(packet);
				}
			});
		}

		if (modemCon != null) {
			// initialize modem connection
			dataConnector = new SafeflyConnector(modemCon, new PacketConnectionOwner() {

				@Override
				public void incommingPacket(PacketInterface packet) {
					System.out.println("Received data in " + getClass().getName());
				}
			});
		}

		// initialize sending thread (Query Integra)
		integraThread = new Thread(new Runnable() {

			@Override
			public void run() {
				sendPendingPackets();
			}
		}, getClass().getName() + " sending thread");

		// initialize thread for informing of current telemetry
		modemThread = new Thread(new Runnable() {

			@Override
			public void run() {
				informData();
			}
		}, getClass().getName() + " GS informing thread");

		// connect
		try {
			if (owner != null || dataConnector.connect()) {
				modemThread.start();
			} else {
				System.err.println(getClass().getName() + "> failed to connect modem.");
			}
		} catch (Exception e) {
			System.err.println(getClass().getName() + "> failed to connect modem.");
			modemThread.start(); // keep sending data
		}

		isConnected = integraConnector.connect();
		// Check and connect special visualization connector
		if(visualizationConnector == null || !visualizationConnector.connect()) {
			// No special visualization connector was set or connection refused
			visualizationConnector = integraConnector;
		}
		integraThread.start();
		
		subsribeGroups();

		return isConnected;
	}
	
	
	protected void subsribeGroups(){
		// subscribe
		if(integraGroups == null){
			LOGGER.warn("No Groups for subscription");
			return;
		}
		
		for (Group group : integraGroups.getGroup()) {
			LOGGER.info("Group ID: " + group.getId());
			LOGGER.info("Channel(s) ID(s): " + group.getChannel());

			this.setGroup(group.getId(), group.getRefreshInterval());

			for (int channel : group.getChannel()) {
				this.addChannel(group.getId(), channel);
			}
		}
	}

	/** Close and disconnect all resources */
	public void dipose() {
		if (integraConnector != null) {
			integraConnector.dispose();
		}
		if(visualizationConnector != null && visualizationConnector != integraConnector) {
			visualizationConnector.dispose();
		}
		integraThread.interrupt();
		modemThread.interrupt();
	}

	// /////////////////////////////////////////////////////////
	// /////// RECEIVE PACKETS
	// /////////////////////////////////////////////////////////

	synchronized protected void processPacket(PacketInterface packet) {
		StringBuffer message = new StringBuffer();
		LOGGER.debug("Processing packet...");
		if (packet instanceof ApplicationPacket) {
			ApplicationPacket appPacket = (ApplicationPacket) packet;
			message.append(String.format("Application packet properties : Type = %d, Command ID = %d \n", appPacket.getCmdType().getValue(), appPacket
					.getCmd_id().getValue()));

			if (appPacket.getCmdType() == Cmd_Type.CMDTYPE_REPLY) {
				// reply -> append reply string to log message
				message.append(String.format("(REPLY)\n"));
				// packet acknowledged, allow sending of next packet
				ackReceived = true;
			} else {
				// not reply -> extract information, append it to the log
				switch (appPacket.cmd_id) {
				case CMDID_RX_DATA:
					// parse data stream
					List<ApplicationData> appData = appPacket.getAppData();
					int i = 0;
					for (ApplicationData appD : appData) {
						if (i++ == 0) {
							message.append(String.format("Group ID = %d, Channel count = %d\n", appD.getGroupOrChannelId(), appData.size() - 1));
							continue;
						}
						ApplicationPacket.CmdRXDataChannel chnlData = (ApplicationPacket.CmdRXDataChannel) appD;
						int channelId = chnlData.getGroupOrChannelId();
						int value = chnlData.getValue();
						parseChannelData(channelId, value);

						message.append(String.format(" Channel# = %d, Value = %d,\n", channelId, value));
					}
					break;

				default:
					throw new RuntimeException("Unknown command id : " + appPacket.cmd_id.getValue());
				}
			}

		} else if(packet instanceof VisualizationPacket){
			VisualizationPacket visPacket = (VisualizationPacket) packet;
			message.append(visPacket.toString());
		}else {
			message.append(String.format("Recieved packet is NOT instance of SafeflyPacket"));
		}

		LOGGER.debug(message.toString());

	}

	/**
	 * Update actual telemetry and other information
	 * 
	 * @param channelId
	 * @param value
	 */
	protected void parseChannelData(int channelId, int value) {
		if (channelId > 4357 && channelId < 4373) {
			// Zaon data
			zaonData.processData(channelId, value);
		} else {
			// Telemetry data
			curState.processData(channelId, value);
		}
	}

	// ////////////////////////////////////////////////////
	// /////// SEND PACKETS
	// ////////////////////////////////////////////////////

	/**
	 * Set a group given by ID and refresh period [ms]
	 * 
	 * @param groupId
	 * @param refreshInt
	 */
	public void setGroup(int groupId, int refreshInt) {
		ApplicationPacket appPacket = new ApplicationPacket(Cmd_ID.CMDID_GRP_SET_PARAM, Cmd_Type.CMDTYPE_REQ, groupId, refreshInt);
		addToSendingQueue(appPacket);
	}

	/**
	 * Add channel to group
	 * 
	 * @param groupId
	 * @param channelId
	 */
	public void addChannel(int groupId, int channelId) {
		ApplicationPacket appPacket = new ApplicationPacket(Cmd_ID.CMDID_CHNL_ADD, Cmd_Type.CMDTYPE_REQ, groupId, channelId);
		addToSendingQueue(appPacket);
	}

	public void addToSendingQueue(PacketInterface packet) {
		sendPacketQueue.add(packet);
	}
	
	/**
	 * Send visualization packet to Integra
	 * @param packet - VisualizationPacket
	 * @return
	 */
	public boolean sendPacketToIntegra(PacketInterface packet) {
		if (visualizationConnector.sendPacket(packet)) {
			LOGGER.info("Packet sent to Integra: " + packet.getApplicationID());
			return true;
		} else {
			LOGGER.warn("Failed to send visualization packet to Integra!");
			return false;
		}
	}

	/**
	 * Send prepared packet waiting in sendPacketQueue
	 */
	protected void sendPendingPackets() {
		while (!Thread.interrupted()) {
			try {
				PacketInterface packet = sendPacketQueue.take();
				ackReceived = false;
				while (!ackReceived) {
					try {
						if (integraConnector.sendPacket(packet)) {
							LOGGER.info("Packet sent to Integra: " + packet.toString());
						} else {
							System.err.println("Failed to send packet to Integra!");
							return;
						}

						Thread.sleep(INTER_PACKET_WAIT_TIME);

					} catch (InterruptedException e) {
						return;
					}
				}
			} catch (InterruptedException e) {
				return;
			}
		}
	}

	/**
	 * Send data via modem
	 * */
	protected void informData() {
		while (!Thread.interrupted()) {
			if (dataConnector != null && dataConnector.isConnected()) {
				Packet packet = new Packet(curState.toPacket());
				// TODO - pass Zaon data to ground?
				if (dataConnector.sendPacket(packet)) {
					LOGGER.debug("Packet sent to GS: " + ByteUtils.byteArrayToHexString(packet.getData()));
				} else {
					System.err.println("Failed to send GS Inform packet!");
					return;
				}
			}
			if (owner != null) {
				owner.newIntegraData(curState);
				owner.newIntegraData(zaonData);
			}

			LOGGER.info(curState);
//			LOGGER.info(zaonData);
			try {
				Thread.sleep(TELEMETRY_SEND_PERIOD);
			} catch (InterruptedException e) {
				return;
			}
		}

	}

	// ////////////////////////////////////////////////////
	// /////// LOGGING
	// ////////////////////////////////////////////////////

	/** Initialize log file */
	protected void initFileLogging() {
		// creating logs directory
		File dir = new File(LOG_DIR);
		if (!dir.exists()) {
			dir.mkdir();
		}
		// creating logfile
		try {
			int i = 1;
			File file = new File(dir + File.separator + LOG_FILE_PREFIX + i + ").txt");
			while (file.exists()) {
				i++;
				file = new File(dir + File.separator + LOG_FILE_PREFIX + i + ").txt");
			}
			logFile = new BufferedWriter(new FileWriter(file));
		} catch (IOException ex1) {
			System.err.println("Cannot create log file: " + ex1);
		}
	}

	/**
	 * Log to file
	 * 
	 * @param log
	 */
	protected void log(String log) {
		LOGGER.info(log);
		if (logFile != null) {
			try {
				String time = DateFormat.getDateTimeInstance().format(new Date());
				logFile.append(time + " : " + log);
				logFile.newLine();
				logFile.flush();
			} catch (IOException ex) {
				ex.printStackTrace();
			}
		}
	}

	// /////////////////////////////////////////////////////////
	// /////////// MAIN METHOD
	// /////////////////////////////////////////////////////////

	private static void initializeLogger() {
		Properties prop = new Properties();
		try {
			InputStream in = ClassLoader.getSystemResourceAsStream("log4j/log4j.properties");
			if (in == null) {
				in = new FileInputStream("log4j/log4j.properties");
			}
			prop.load(in);
		} catch (Exception ex) {
			ex.printStackTrace();
		}
		PropertyConfigurator.configure(prop);
	}

	/** prints out help text for main() method */
	private static void printHelp() {
		System.out.println();
		System.out.println("============================< Gumstix Client Help >============================");
		System.out.println();
		System.out.println("NAME");
		System.out.println("    GumstixClient.jar --Gumstix client for Integra-Gumstix communication");
		System.out.println();
		System.out.println("ARGUMENTS");
		System.out.println("    [no arguments | -h | -help]");
		System.out.println("                --print out this help");
		System.out.println();
		System.out.println("    [-d | -default]  ");
		System.out.println("                --load configuration from default location");
		System.out.println("                  xml/serial_port.xml");
		System.out.println("                  xml/integra_channels.xml");
		System.out.println("                  xml/modem_channels.xml");
		System.out.println();
		System.out.println("    [-portxmlfile | -portfile | -pf ] filepath    ");
		System.out.println("                --set path to Port Xml configuration file");
		System.out.println();
		System.out.println("    [-channelsxmlfile | -channelsfile | -chf] filepath    ");
		System.out.println("                --set path to groups/channels Xml configuration file");
		System.out.println();
		System.out.println("    [-modemxmlfile | -modemfile | -mf] filepath    ");
		System.out.println("                --set path to modem Port Xml configuration file");
		System.out.println();
		System.out.println("DESCRIPTION");
		System.out.println("    Sends groups requests to integra and logs incoming datastream.");
		System.out.println("    Port and groups configuration is read from external xml files.");
	}

	/**
	 * Parse input
	 * 
	 * @param arguments
	 * @return String array - [0] integraPortXmlLocation; [1]
	 *         channelsXmlLocation; [2] modemPortXmlLocation
	 */
	private static String[] parseInputArguments(String[] arguments) {

		String portXmlLocation = DEFAULT_INTEGRA_PORT_XML_LOCATION;
		String channelsXmlLocation = DEFAULT_CHANNELS_XML_LOCATION;
		String modemXmlLocation = DEFAULT_MODEM_PORT_XML_LOCATION;

		// parsing input arguments
		if (arguments.length == 0) {
			printHelp();
		}

		for (int i = 0; i < arguments.length; i++) {
			if (arguments[i].equals("-portxmlfile") | arguments[i].equals("-portfile") | arguments[i].equals("-pf")) {
				portXmlLocation = nextArg(i++, arguments);
			} else if (arguments[i].equals("-channelsxmlfile") | arguments[i].equals("-channelsfile") | arguments[i].equals("-chf")) {
				channelsXmlLocation = nextArg(i++, arguments);
			} else if (arguments[i].equals("-modemxmlfile") | arguments[i].equals("-modemfile") | arguments[i].equals("-mf")) {
				modemXmlLocation = nextArg(i++, arguments);
			}
		}
		String[] result = { portXmlLocation, channelsXmlLocation, modemXmlLocation };
		return result;
	}

	/**
	 * Read and return next argument
	 * 
	 * @param i
	 * @param args
	 * @return
	 */
	private static String nextArg(int i, String[] args) {
		if (i + 1 == args.length) {
			System.err.println("missing parameter after marker " + args[i]);
			System.exit(-1);
		}
		return args[i + 1];
	}

	public static void main(String[] args) throws InterruptedException {
		initializeLogger();

		LOGGER.info("Call MAIN method");
		// parse input arguments
		String[] parsingResult = parseInputArguments(args);
		String portXmlLocation = parsingResult[0];
		String channelsXmlLocation = parsingResult[1];
		String modemXmlLocation = parsingResult[2];

		// load group and channel settings
		SerialPortSettings integraCom = (SerialPortSettings) XmlLoader.loadXml(portXmlLocation, SerialPortSettings.class);
		Groups gr = (Groups) XmlLoader.loadXml(channelsXmlLocation, Groups.class);
		SerialPortSettings modemCom = (SerialPortSettings) XmlLoader.loadXml(modemXmlLocation, SerialPortSettings.class);

//		IntegraCommunicator integraComm = new IntegraCommunicator(integraCom, modemCom);
		IntegraCommunicator integraComm = new IntegraCommunicator(integraCom, null);
		integraComm.connect();

		// subscribe
		for (Group group : gr.getGroup()) {
			System.out.println("Group ID: " + group.getId());
			System.out.println("Channel(s) ID(s): " + group.getChannel());

			integraComm.setGroup(group.getId(), group.getRefreshInterval());

			for (int channel : group.getChannel()) {
				integraComm.addChannel(group.getId(), channel);
			}
		}

		Scanner sc = new Scanner(System.in);
		while (!sc.nextLine().equalsIgnoreCase("Q")) {
			System.out.println("Type Q to exit.");
		}
		sc.close();
		integraComm.dipose();
	}

}
