package atc.safefly.communication;

import java.util.ArrayList;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;

import atc.ontology.gpsflightplan.GpsState;
import javax.vecmath.Point3d;
import aglobex.simulation.global.ConversionConstants;
import atc.safefly.communication.common.SafeflyConnector;
import atc.safefly.communication.common.packet.ApplicationPacket;
import atc.safefly.communication.common.packet.PacketConnectionOwner;
import atc.safefly.communication.common.packet.PacketInterface;
import atc.safefly.communication.common.packet.ApplicationPacket.CmdAddChannel;
import atc.safefly.communication.common.packet.ApplicationPacket.CmdDeactGroup;
import atc.safefly.communication.common.packet.ApplicationPacket.CmdGroupSet;
import atc.safefly.communication.common.packet.VisualizationPacket;
import atc.safefly.communication.common.serial.SerialConnector;
import atc.safefly.communication.common.serial.UniversalSerialConnector;
import atc.safefly.communication.utils.ApplicationData;
import atc.safefly.utils.ByteUtils;
import atc.safefly.utils.SafeflyConstants;
import atc.safefly.utils.SafeflyConstants.Cmd_ID;
import atc.safefly.utils.SafeflyConstants.Cmd_Type;
import atc.safefly.utils.SafeflyUtils;
import atc.safefly.wptsgenerator.GeneratedWaypoints;
import atc.ontology.gpsflightplan.math.GpsPosition;

/**
 * Class that simulates Integra Glass Cockpit data channel. Uses Measure
 * Protocol.
 * 
 */
public class IntegraSim {

	SafeflyConnector connector = null;

	private static final String DEFAULT_PORT = "COM14";

	private int baudRate = 115200;
	private GroupThread[] grpArray = new GroupThread[8];
	private static final int GRP_ID_MIN = 0;
	private static final int GRP_ID_MAX = 7;
	private Hashtable<Integer, Integer> channelData = new Hashtable<Integer, Integer>();

	// ///////////////////////
	// GPS WAYPOINT DATA
	// ///////////////////////

	/**
	 * current GPS waypoint - (longitude, latitude, altitude) in radians and
	 * meters
	 */
	private Point3d curGPSWaypoint;

	/** 5char name of the current waypoint */
	private String curWptName;
	private int wptId = 0;
	
	
	public IntegraSim(String port) {
		SerialConnector sercom = new UniversalSerialConnector(port, baudRate);
		connector = new SafeflyConnector(sercom, new PacketConnectionOwner() {
			
			@Override
			public void incommingPacket(PacketInterface packet) {
				processPacket(packet);
			}
		});
		if(!connector.connect()) {
			System.err.println("Failed to open serial connector");
		}
	}

	public IntegraSim() {
		this(DEFAULT_PORT);
	}

	/**
	 * Set data for channel
	 * 
	 * @param chnlId
	 * @param chnlVal
	 */
	public void setChannelData(int chnlId, int chnlVal) {
		this.channelData.put(chnlId, chnlVal);
	}

	/**
	 * Set data from {@link GpsState}
	 * 
	 * @param state
	 */
	public void setChannelData(GpsState state) {
		setChannelData(SafeflyConstants.CH_INTEGRA_ID, state.id);
		setChannelData(SafeflyConstants.CH_GPS_ALTITUDE, (int) (state.getAltitudeM() * 10));
		setChannelData(SafeflyConstants.CH_GPS_LATITUDE, (int) (state.getLatitudeDeg() * 100000));
		setChannelData(SafeflyConstants.CH_GPS_LONGITUDE, (int) (state.getLongitudeDeg() * 100000));

		setChannelData(SafeflyConstants.CH_PITCH, (int) (state.getPitchDeg() * 100));
		setChannelData(SafeflyConstants.CH_ROLL, (int) (state.getBankDeg() * 100));
		setChannelData(SafeflyConstants.CH_AZIMUT, (int) (state.getHeadingDeg() * 100));

		setChannelData(SafeflyConstants.CH_SPEED_TAS, (int) (state.velocityTasMS * ConversionConstants.MS_TO_KMH * 10));
		setChannelData(SafeflyConstants.CH_GPS_SPEED_GROUND, (int) (state.velocityTgsMS * ConversionConstants.MS_TO_KMH * 10));

		int[] timeData = SafeflyUtils.getIntegraDateAndTimeFromMS(state.timestampMs);
		setChannelData(SafeflyConstants.CH_DATE, timeData[0]);
		setChannelData(SafeflyConstants.CH_TIME, timeData[1]);

		// waypoint data:
		if (curGPSWaypoint != null) {
			double[] dstBearing = SafeflyUtils.getDistAndBearingFromTwoGpsPoints(state.getLongitudeDeg() * ConversionConstants.DEG_TO_RAD,
					state.getLatitudeDeg() * ConversionConstants.DEG_TO_RAD, curGPSWaypoint.x, curGPSWaypoint.y, (state.getAltitudeM() + curGPSWaypoint.z) / 2);

			setChannelData(SafeflyConstants.CH_WPT_DIST, (int) (dstBearing[0] / 100)); // *10/1000
			setChannelData(SafeflyConstants.CH_WPT_BEARING, (int) (dstBearing[1] * ConversionConstants.RAD_TO_DEG * 10));
			int wptIdT;
			try{
				wptIdT = Integer.parseInt(curWptName);
			}
			catch(NumberFormatException e){
				wptIdT = wptId;
			}
			setChannelData(SafeflyConstants.CH_WPT_ID, wptIdT);
		}

	}

	private boolean isSupportedGroup(int grpId) {
		if (grpId >= GRP_ID_MIN & grpId <= GRP_ID_MAX) {
			return true;
		} else {
			return false;
		}
	}

	public void groupSet(int grpId, int refreshInterval) {
		if (isSupportedGroup(grpId)) {
			if (this.grpArray[grpId] != null) {
				this.grpArray[grpId].setGrpId(grpId);
				this.grpArray[grpId].setRefreshInterval(refreshInterval);
				this.grpArray[grpId].activate();
			} else {
				this.grpArray[grpId] = new GroupThread(grpId, refreshInterval, channelData, connector);
				this.grpArray[grpId].start();
			}
		} else {
			throw new RuntimeException("Cannot set unsupported group. Group id must be in interval of " + GRP_ID_MIN + ".." + GRP_ID_MAX);
		}
	}

	public void groupDeact(int grpId, int flags) {
		if (isSupportedGroup(grpId)) {
			if (this.grpArray[grpId] != null) {
				this.grpArray[grpId].deactivate(flags);
			} else {
				throw new RuntimeException("Cannot deactivate group " + grpId + " have not been created yet.");
			}
		} else {
			throw new RuntimeException("Cannot deactivate unsupported group. Group id must be in interval of " + GRP_ID_MIN + ".." + GRP_ID_MAX);
		}
	}

	public void channelAdd(int grpId, int chnlId) {
		if (isSupportedGroup(grpId)) {
			if (this.grpArray[grpId] != null) {
				this.grpArray[grpId].channelAdd(chnlId);
			} else {
				throw new RuntimeException("Cannot add channel to group. Group " + grpId + " have not been created yet.");
			}
		} else {
			throw new RuntimeException("Cannot add channel to unsupported group. Group id must be in interval of " + GRP_ID_MIN + ".." + GRP_ID_MAX);
		}
	}

	/**
	 * Update actual target waypoint and next heading
	 * 
	 * @param wpts
	 */
	public void waypointsReceived(GeneratedWaypoints wpts) {
		GpsPosition pos = wpts.getCurrWpt().position;
		curGPSWaypoint = new Point3d(pos.getLongitudeDeg() * ConversionConstants.DEG_TO_RAD, pos.getLatitudeDeg() * ConversionConstants.DEG_TO_RAD,
				pos.getAltitudeM());

		String wptNameFull = wpts.getCurrWpt().name;
		String wptName = SafeflyUtils.fixedLenthString(wptNameFull, 5);
		if(!wptName.equals(curWptName)){
			curWptName = wptName;
			wptId++;
		}
		

	}

	// ////////////////////////////////////////////////////////
	// ///// CLASS GROUPTHREAD
	// ////////////////////////////////////////////////////////

	private class GroupThread extends Thread {

		private int grpId;
		private int refreshInterval;
		protected boolean active;
		private HashSet<Integer> channelsInGroup = new HashSet<Integer>();
		private Hashtable<Integer, Integer> channelData = new Hashtable<Integer, Integer>();
		SafeflyConnector connector;
		private ApplicationPacket appPacket;

		private class Lock {
			private boolean isLocked = false;

			public synchronized void lock() throws InterruptedException {
				while (isLocked) {
					wait();
				}
				isLocked = true;
			}

			public synchronized void unlock() {
				isLocked = false;
				notify();
			}
		}

		private Lock lock = new Lock();

		public void setGrpId(int grpId) {
			this.grpId = grpId;
		}

		public void setRefreshInterval(int refreshInterval) {
			this.refreshInterval = refreshInterval * 10;
		}

		/**
		 * creates thread to send required group data
		 * 
		 * @param grpId
		 *            - ID of the group
		 * @param refreshInterval
		 *            - interval in which will be data send [ms * 10]
		 * @param channelData
		 *            - reference to data source
		 * @param connector
		 *            - reference to related UAVConnector
		 */
		public GroupThread(int grpId, int refreshInterval, Hashtable<Integer, Integer> channelData, SafeflyConnector connector) {
			this.grpId = grpId;
			this.refreshInterval = refreshInterval * 10;
			this.channelData = channelData;
			this.active = true;
			this.connector = connector;
		}

		public int getGrpId() {
			return grpId;
		}

		public void deactivate(int flags) {
			if (flags == 1) {
				this.channelsInGroup.clear();
			}
			this.active = false;
		}

		public void activate() {
			this.active = true;
		}

		public void channelAdd(int chnlId) {
			try {
				lock.lock();
			} catch (InterruptedException e) {
				e.printStackTrace();
			}
			this.channelsInGroup.add(chnlId);
			lock.unlock();
		}

		@Override
		public void run() {
			ArrayList<Integer> dataList = new ArrayList<Integer>();
			while (true) {
				try {
					Thread.sleep(this.refreshInterval);
					if (active) {
						lock.lock();
						Iterator<Integer> it = channelsInGroup.iterator();
						int channel;
						int data;
						dataList.clear();
						dataList.add((Integer) this.grpId);
						dataList.add((Integer) this.channelsInGroup.size());
						while (it.hasNext()) {
							channel = it.next();
							if (channelData.containsKey(channel)) {
								data = channelData.get(channel);
							} else {
								data = -9999;
							}
							dataList.add((Integer) channel);
							dataList.add((Integer) data);
						}
						this.appPacket = new ApplicationPacket(Cmd_ID.CMDID_RX_DATA, Cmd_Type.CMDTYPE_REQ, convertIntegers(dataList));
						lock.unlock();
						this.connector.sendPacket(this.appPacket);

					}
				} catch (InterruptedException e) {
					System.err.println("GroupThread interupted early grpId = " + this.grpId);
					e.printStackTrace();
				}
			}
		}

	}

	/**
	 * Process arrived packet
	 * 
	 * @param packet
	 * @param connector
	 */
	private void processPacket(PacketInterface packet) {
		if (packet instanceof ApplicationPacket) {
			ApplicationPacket appPacket = (ApplicationPacket) packet;
			System.out.println("Recieved packet is instance of ApplicationPacket");

			Cmd_ID cmdId = appPacket.getCmd_id();
			int grpId;
			int chnlId;
			int refreshInterval;
			short flags;
			
			ApplicationPacket appPacketReply = null;

			ApplicationData appData = appPacket.getAppData().get(0);

			System.out.println("recieved packet with cmd id = " + cmdId);
			switch (cmdId) {
			case CMDID_GRP_SET_PARAM: // group set request
				grpId = ((CmdGroupSet) appData).getGroupOrChannelId();
				refreshInterval = ((CmdGroupSet) appData).getRefreshInt();

				appPacketReply = new ApplicationPacket(Cmd_ID.CMDID_GRP_SET_PARAM, Cmd_Type.CMDTYPE_REPLY, 0);
				// grpId = ByteUtils.bytesToShort(new byte[] {appPacketData[2],
				// 0 });
				// refreshInterval = ByteUtils.bytesToInt(new byte[]
				// {appPacketData[3], appPacketData[4] });
				this.groupSet(grpId, refreshInterval);
				break;

			case CMDID_GRP_DEACT: // group deactivate request
				grpId = ((CmdDeactGroup) appData).getGroupOrChannelId();
				flags = ((CmdDeactGroup) appData).getFlags();

				appPacketReply = new ApplicationPacket(Cmd_ID.CMDID_GRP_DEACT, Cmd_Type.CMDTYPE_REPLY, 0);
				// grpId = ByteUtils.bytesToShort(new byte[] {appPacketData[2],
				// 0 });
				// flags = ByteUtils.bytesToInt(new byte[] {appPacketData[3],
				// 0});
				this.groupDeact(grpId, flags);
				break;

			case CMDID_CHNL_ADD: // add channel to group request
				grpId = ((CmdAddChannel) appData).getGroupOrChannelId();
				chnlId = ((CmdAddChannel) appData).getChnlId();

				appPacketReply = new ApplicationPacket(Cmd_ID.CMDID_CHNL_ADD, Cmd_Type.CMDTYPE_REPLY, 0);
				// grpId = ByteUtils.bytesToShort(new byte[] {appPacketData[2],
				// 0});
				// chnlId = ByteUtils.bytesToInt(new byte[] {appPacketData[3],
				// appPacketData[4] });
				this.channelAdd(grpId, chnlId);
				break;

			case CMDID_RX_DATA: // incomming data stream
				appPacketReply = new ApplicationPacket(Cmd_ID.CMDID_RX_DATA, Cmd_Type.CMDTYPE_REPLY, 0);
				System.err.println("Handling data stream is not implemented.");
				break;
			default:
				throw new RuntimeException("Unknown command id : " + cmdId);
			}

			System.out.println("packet Reply " + ByteUtils.byteArrayToHexString(appPacketReply.serialize()));

			if (this.connector.sendPacket(appPacketReply)) {
				System.out.println("Reply send suscesfull");
			} else {
				System.out.println("Reply send failed");
			}

		} 
		else if(packet instanceof VisualizationPacket){
			System.err.println("IntegraSim: Received Visualization Packet that I cannot parse ! ");
		}
		else {
			throw new RuntimeException("Recieved packet is NOT instance of SafeflyPacket");
		}
		System.out.println();
	}

	public static synchronized int[] convertIntegers(ArrayList<Integer> integers) {
		int[] ret = new int[integers.size()];
		Iterator<Integer> iterator = integers.iterator();
		for (int i = 0; i < ret.length; i++) {
			ret[i] = iterator.next().intValue();
		}
		return ret;
	}

	public static void main(String[] args) {
		System.out.println("IntegraSim running...");

		IntegraSim rH = new IntegraSim();

		// rH.groupSet(2, 500);
		// rH.channelAdd(2, 1280);
		// rH.channelAdd(2, 1281);
		// rH.channelAdd(2, 1282);
		// rH.channelAdd(2, 1283);
		// rH.channelAdd(2, 1284);
		// rH.setChannelData(1281, 1281);

		// Groups gr = (Groups) XmlLoader.loadXml("xml/integra_channels.xml",
		// Groups.class);
		// for (Group g : gr.getGroup()) {
		// System.out.println(g.getChannel());
		//
		// rH.groupSet(g.getId(), g.getRefreshInterval());
		// Iterator<Integer> iterator = g.getChannel().iterator();
		// while (iterator.hasNext()) {
		// rH.channelAdd(g.getId(), iterator.next().intValue());
		// }
		// }
	}

}
