package atc.safefly.communication;

import java.io.IOException;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.nio.ByteBuffer;

import atc.safefly.communication.common.SafeflyConnector;
import atc.safefly.communication.common.packet.Packet;
import atc.safefly.communication.common.packet.PacketConnectionOwner;
import atc.safefly.communication.common.packet.PacketConnector;
import atc.safefly.communication.common.packet.PacketInterface;
import atc.safefly.communication.common.serial.SerialConnector;
import atc.safefly.communication.common.serial.ServerSocketConnector;
import atc.safefly.communication.common.serial.SocketSerialConnector;
import atc.safefly.communication.modem.MicrohardDiagnosticPacketConnector;
import atc.safefly.communication.utils.RssiInfo;
import atc.safefly.communication.utils.WGConnectionOwner;
import atc.safefly.utils.SafeflyTelemetry;

/**
 * Connector for TCP/IP communication with WiGnome via wi-fi
 * 
 * @author Selecky
 *
 */
public class WGConnector {
	
	
	protected WGConnectionOwner owner;

	protected SafeflyTelemetry telem = new SafeflyTelemetry();
	
    protected PacketConnector dataConn;
    
    protected boolean isDataConnected = false;
    
    
    //////////////////////////////////////////////
    //////////  MODEM DIAGNOSTICS
    //////////////////////////////////////////////

    protected MicrohardDiagnosticPacketConnector diagConn;
    
    /** Requested parametrs: RSSI */
 	public static final int[] parametrsID = new int[] { 105 };
 	
 	/** diagnostics frequency in ms */
 	protected final static int DIAG_FREQ_MS = 2000;

    protected boolean isDiagConnected = false;

    

    /**
     * Start WiGnome connector
     * @param owner
     * @param address
     * @param port
     * @param isServer - false for connection to WiGnome server, true for WiGnome emulation, 
     * 					when the server socket is created, it blocks the current thread and waits for connection
     */
    public WGConnector(final WGConnectionOwner owner, String address, int port, boolean isServer) {
    	this.owner = owner;
    	SerialConnector serCon;
    	
    	try {
    		if(isServer){
    			serCon = new ServerSocketConnector(port);
    		} else {
    			serCon = new SocketSerialConnector(InetAddress.getByName(address), port);
    		}
    		
    		dataConn = new SafeflyConnector(serCon, new PacketConnectionOwner() {
				
				@Override
				public void incommingPacket(PacketInterface packet) {
					SafeflyTelemetry telem = new SafeflyTelemetry();
					telem.fromPacket(packet.getData());
					owner.incommingMessage(telem);
				}
			});
    		
		} catch (UnknownHostException e) {
			e.printStackTrace();
		}
    	
    }
    
    
    public boolean connect() throws IOException {
    	if (dataConn.connect()) {
    		
    		isDataConnected = true;
    		return true;
    	}
    	return false;
    }
    
    
    /**
     * Connect to diagnostic port of the MicroHard modem
     * @param address - WiGnome address - default 192.168.2.155
     * @param port - WiGnome port for diagnostic 
     * @return true if connection was successful 
     */
    public boolean connectToDiagPort(String address, int port){
    	try {
    		SerialConnector conn = new SocketSerialConnector(InetAddress.getByName(address), port);
			diagConn = new MicrohardDiagnosticPacketConnector(conn);
			isDiagConnected = diagConn.connect();
			return isDiagConnected;
		} catch (UnknownHostException e) {
			e.printStackTrace();
		} 
    	return false;
    }
    
    /**
     * Start diagnostics thread 
     */
    public void startDiagnostics(){
    	Thread rssiRead = new Thread(new Runnable() {

			@Override
			public void run() {
				while (!Thread.interrupted()) {
					// get active slave modems
					int[] slaveUAs = getSlaves();
					
					// get master RSSI
					byte[] masterLog = diagConn.getParametrsByID(parametrsID, 1);
					int masterRSSI = masterLog[4] & 0xFF;
					
					// get RSSI from slaves
					for(int slaveUA : slaveUAs){
						byte[] slaveLog = diagConn.getParametrsByID(parametrsID, slaveUA);
						int slaveRSSI = 0;
						if(slaveLog != null){
							slaveRSSI = slaveLog[4] & 0xFF;
						}
						RssiInfo rssi = new RssiInfo(masterRSSI, slaveRSSI, slaveUA);
						owner.receiveDiagnostics(rssi);
					}
					
					try {
						Thread.sleep(DIAG_FREQ_MS);
					} catch (InterruptedException e) {
						return;
					}
				}
			}
		});
    	rssiRead.start();
    }

    
    /** Get array of active slave modem addresses */
    protected int[] getSlaves(){
    	byte[] config = diagConn.getNetworkConfiguration();
		while (config == null || config.length < 9) {
			try {
				Thread.sleep(DIAG_FREQ_MS);
			} catch (InterruptedException e) {
				return null;
			}
			config = diagConn.getNetworkConfiguration();
		}
		ByteBuffer bb = ByteBuffer.allocate(2);
		int[] slaveUAs = new int[(config.length - 9) / 2];
		for (int i = 7; i < config.length - 2; i += 2) {
			bb.put(0, config[i]);
			bb.put(1, config[i + 1]);
			slaveUAs[(i - 7) / 2] = bb.getShort(0) & 0xFFFF;
		}
		return slaveUAs;
    }
    

    /**
     * Send object via socket
     * @param pPacket - so far only {@link SafeflyTelemetry} is allowed to be sent 
     * @return
     */
    public boolean sendViaSocket(Object pPacket) {
    	if (isDataConnected) {
    		if(pPacket instanceof SafeflyTelemetry){
    			Packet packet = new Packet(((SafeflyTelemetry) pPacket).toPacket());
    			dataConn.sendPacket(packet);
    			return true;
    		}
    	}
    	return false;
    }


    public void dispose() {
    	if (dataConn != null) {
    		dataConn.dispose();
    	}
    }

}
