package atc.safefly.communication.common;

import java.io.IOException;
import java.net.SocketTimeoutException;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;

import org.apache.log4j.Logger;

import atc.safefly.communication.common.packet.AddressedPacket;
import atc.safefly.communication.common.packet.ApplicationPacket;
import atc.safefly.communication.common.packet.FragmentPacket;
import atc.safefly.communication.common.packet.Packet;
import atc.safefly.communication.common.packet.PacketConnectionOwner;
import atc.safefly.communication.common.packet.PacketConnector;
import atc.safefly.communication.common.packet.PacketFactory;
import atc.safefly.communication.common.packet.PacketInterface;
import atc.safefly.communication.common.packet.TraceRoutePacket;
import atc.safefly.communication.common.packet.VisualizationPacket;
import atc.safefly.communication.common.serial.SerialConnector;
import atc.safefly.plan.SafeflyFlightPlan;
import atc.safefly.utils.ByteUtils;
import atc.safefly.utils.SafeflyConstants;

public class SafeflyConnector extends PacketConnector {
	
	//Initialize Packet classes
	static {
		new Packet();
		new AddressedPacket();
		new FragmentPacket();
		new ApplicationPacket();
		new VisualizationPacket();
		new TraceRoutePacket();
	}
	
	private static final Logger LOGGER = Logger.getLogger(SafeflyConnector.class);

	public static final byte START = 0x11;

	private static final int BUFFER_SIZE = 10240;

	private static final int MAX_PACKET_SIZE = 255;
	
	private boolean verbose = false;
	
	public SafeflyConnector(SerialConnector subConnector, PacketConnectionOwner owner, boolean verbose) {
		super(subConnector, owner);
		this.verbose = verbose;
	}
	
	public SafeflyConnector(SerialConnector subConnector, PacketConnectionOwner owner) {
		super(subConnector, owner);
	}

	public SafeflyConnector(SerialConnector subConnector, boolean verbose) {
		super(subConnector, null);
		this.verbose = verbose;
	}
	
	public SafeflyConnector(SerialConnector subConnector) {
		super(subConnector, null);
	}

	@Override
	public synchronized boolean sendPacket(PacketInterface packet) {
		if(verbose) {
//			LOGGER.info(">>>"+packet);
		}
		
		// Currently not used
		byte flags = 0x00;

		if (isConnected()) {

			byte[] sPacket = packet.serialize();
			// Buffer for 8B header with headCRC + packet size + 2B CRC check
			ByteBuffer bb = ByteBuffer.allocate(8 + sPacket.length + 2);
			bb.order(ByteOrder.LITTLE_ENDIAN);
			short headCRC = 0;
			/****************** HEADER ******************************/
			bb.put(START);
			bb.put(flags);
			bb.putShort((short) packet.getApplicationID());
			bb.putShort((short) sPacket.length);
			headCRC = computeCRC(bb, 6);
			bb.putShort(headCRC);
			/****************** BODY ********************************/
			bb.put(sPacket);
			bb.putShort(computeCRC(bb, 8 + sPacket.length));
			/********************************************************/

			try {
				if(verbose) {
//					LOGGER.info(">>>"+ByteUtils.byteArrayToHexString(bb.array()));
				}
//				System.err.println("Length> "+bb.array().length);
				out.write(bb.array());
				out.flush();
				StringBuilder sb = new StringBuilder();
				sb.append(SafeflyConnector.class.getName()).append('\t').append(subConnector.getPortName()).append('\n');
				sb.append("Packet sent: ").append(packet.getClass().getName()).append('\n');
				sb.append(packet.toString()).append('\n');
				sb.append("Bytes: ").append(ByteUtils.byteArrayToHexString(bb.array()));
				LOGGER.info(sb.toString());
				return true;
			} catch (IOException e) {
				dispose();
				e.printStackTrace();
			}
		}
		
		else if(verbose){
			LOGGER.warn("SafeflyConnector: cannot send packet - not connected");
		}
		return false;
	}

	@Override
	protected void readPacket() {
		ByteBuffer bb = ByteBuffer.allocateDirect(4*BUFFER_SIZE);
		bb.order(ByteOrder.LITTLE_ENDIAN);
		bb.limit(0);
		byte[] buffer = new byte[BUFFER_SIZE];

		byte flags = 0;
		short applicationID = 0;
		int dataLength = -1;
		
		while (!Thread.interrupted()) {
			/****************** Read data into Buffer ****************************/
			try {
				int size = in.read(buffer);
				if(size > 0) {
//					System.out.println("<<<"+ByteUtils.byteArrayToHexString(buffer, 0, size));
	
					// Add data at the end of ByteBuffer
					int curLimit = bb.limit();
					bb.limit(curLimit+size);
					bb.position(curLimit);
					bb.put(buffer,0,size);
					bb.position(curLimit);
				}else if (size > -1) {
					LOGGER.error(getClass().getSimpleName()+": No data read from input stream.");
				}else {
					System.err.println(getClass().getSimpleName()+": End-Of-Stream!");
					LOGGER.error(getClass().getSimpleName()+": End-Of-Stream!");
					dispose();
					return;
				}
			} catch(SocketTimeoutException e) {
				continue;
			} catch (IOException e) {
				System.err.println(getClass().getSimpleName()+": Failed to read data!");
				dispose();
				return;
			}
			/**********************************************************************/
			
			while (bb.hasRemaining()) {
				/***************** Read start *************************************/
				while (bb.get(0) != START && bb.hasRemaining()) {
					if (bb.get() == START) {
						bb.position(bb.position()-1);
						bb.compact().flip();
					}
				}
				
				/**************** Check header ***********************************/
				if(bb.get(0) == START && bb.limit() > 8) {
					//Header received
					if(dataLength >= 0) {
						//Process packet
						if(bb.limit() >= (10 + dataLength)) {
							//Check packet CRC
							if(bb.getShort(dataLength+8) == computeCRC(bb, 8 + dataLength)){
								// Process correct packet
									try {
										// TODO / remove this debug
//										byte[] temp = new byte[dataLength+10];
//										bb.position(0);
//										bb.get(temp);
//										System.out.println("Received: "+ByteUtils.byteArrayToHexString(temp));
										
										bb.position(8);
										PacketInterface packet = PacketFactory.getPacket(applicationID);
										packet.deserialize(bb, dataLength);
										bb.order(ByteOrder.LITTLE_ENDIAN);
										if(verbose) {
//											System.out.println("<<<"+packet);
//											LOGGER.info("<<<"+packet);
										}
										packetQueue.put(packet);
										bb.position(10+dataLength);
										bb.compact().flip();
										dataLength = -1;
									} catch (InterruptedException e) {
										return;
									}
							}else {
								// Dismiss corrupted packet
								System.err.println(getClass().getSimpleName()+": Packet CRC corrupted!");
								//Skip current packet data
								bb.position(10 + dataLength);
								bb.compact().flip();
								dataLength = -1;
							}
						}else {
							break;
						}
					}else {
						//Check and read header
						if(bb.getShort(6) == computeCRC(bb, 6)){
							flags = bb.get(1);
							applicationID = bb.getShort(2);
							dataLength = bb.getShort(4) & 0xFFFF;
						}else {
							System.err.println(getClass().getSimpleName()+": Header CRC corrupted!");
							LOGGER.error("Header CRC corrupted!");
							// Reset current corrupted START byte
							bb.put(0,(byte)0);
							// Continue to find new START byte with correct header
						}
					}
				}else {
					break;
				}
			}
		}
	}
	
	private short computeCRC(ByteBuffer bb, int length) {
		int crc = 0x0000;
		for (int i = 0; i < length; i++) {
			crc = (crc >>> 8) ^ SafeflyConstants.CRC_LOOKUP_TABLE[(crc ^ bb.get(i)) & 0xff];
		}
		return (short)crc;
	}
	
	@Override
	public int getMaxPacketSize() {
		return MAX_PACKET_SIZE;
	}
}
