package atc.safefly.communication.common.packet;

import java.io.UnsupportedEncodingException;
import java.nio.ByteBuffer;

import aglobe.ontology.Message;
import aglobex.entitycontrol.deconfliction.cooperative.utilitybased.GeneratedPlans;
import atc.ontology.gpsflightplan.GpsFlightPlan;
import atc.safefly.plan.SafeflyFlightPlan;
import atc.safefly.plan.SafeflyGeneratedPlans;
import atc.safefly.utils.ByteUtils;


/**
 * Class for simplified serialization and de-serialization of Aglobe messages
 * @author Selecky
 *
 */
public class AglobeMsgContentSerializer {
	
	private final static byte CONTENT_GENERAL = 1;
	private final static byte CONTENT_GPS_FLIGHT_PLAN = 2;
	private final static byte CONTENT_GENERATED_PLANS = 3; 
	
		
	//---------------------------------------------------------------
	//----------    Serialization
	//---------------------------------------------------------------
	
	/**
	 * Serialize Aglobe Message
	 * @param m
	 * @return
	 */
	public static byte[] serialize(Message m) {
		String protocol = m.getProtocol();
		String performative = m.getPerformative();
		String reason = m.getReason();
		String conversationId = m.getConversationID();
		Object content = m.getContent();
		
		// prepare parameters to write out
		byte protSize = 0;
		byte perfSize = 0;
		byte reasonSize = 0;
		byte cIdSize = 0;
		
		if(protocol != null){
			protSize = (byte)protocol.length();
		}
		if(performative != null){
			perfSize = (byte)performative.length();
		}
		if(reason != null){
			reasonSize = (byte)reason.length();
		}
		if(conversationId != null){
			cIdSize = (byte)conversationId.length();
		}
		
		// serialize Message content 
		byte[] contB = null;
		byte contentType = -1;
		
		if(content instanceof GpsFlightPlan){
			SafeflyFlightPlan sfPlan = new SafeflyFlightPlan((GpsFlightPlan)content);
			contB = ByteUtils.object2bytes(sfPlan);
			contentType = CONTENT_GPS_FLIGHT_PLAN;
		}
		else if (content instanceof GeneratedPlans){
			SafeflyGeneratedPlans gPlans = new SafeflyGeneratedPlans((GeneratedPlans) content);
			contB = ByteUtils.object2bytes(gPlans);
			contentType = CONTENT_GENERATED_PLANS;
		}
		else {
			contB = ByteUtils.object2bytes(content);
			contentType = CONTENT_GENERAL;
		}
		

		// write into ByteBuffer 
		int contSize = contB.length;
		
		ByteBuffer bb = ByteBuffer.allocate(perfSize + protSize + reasonSize + cIdSize + contSize + 9);  // 8 = 5*byte(sizes,type) +1*int
		try {
			bb.put(protSize);
			if(protocol != null) {
				bb.put(protocol.getBytes("UTF-8"));
			}
			bb.put(perfSize);
			if(performative != null){
				bb.put(performative.getBytes("UTF-8"));
			}
			bb.put(reasonSize);
			if(reason != null){
				bb.put(reason.getBytes("UTF-8"));
			}
			bb.put(cIdSize);
			if(conversationId != null){
				bb.put(conversationId.getBytes("UTF-8"));
			}
			
			bb.putInt(contSize);
			bb.put(contentType);
			bb.put(contB);
			
			return bb.array();
			
		} catch (UnsupportedEncodingException e) {
			e.printStackTrace();
		}

		return null;
	}

	
	
	//---------------------------------------------------------------
	//----------    Deserialization
	//---------------------------------------------------------------
	/**
	 * Deserialize Aglobe Message
	 * @param data
	 * @return
	 */
	public static Message deserialize(byte[] data) {
		String protocol = null;
		String performative = null;
		String reason = null;
		String conversationId = null;
		Object content = null;
		
		ByteBuffer bb = ByteBuffer.wrap(data);
		
		// read protocol
		int size = bb.get();
		protocol = ByteUtils.readString(bb, size);
		
		// read performative
		size = bb.get();
		performative = ByteUtils.readString(bb, size);
		
		// read reason
		size = bb.get();
		reason = ByteUtils.readString(bb, size);
		
		// read conversationId
		size = bb.get();
		conversationId = ByteUtils.readString(bb, size);
		
		// read content
		size = bb.getInt();
		byte[] contB = new byte[size];
		byte contentType = bb.get();
		bb.get(contB);
		try {
			// read data
			Object readData = ByteUtils.bytes2object(contB);
			
			switch (contentType) {
			case CONTENT_GPS_FLIGHT_PLAN:
				// get SafeflyFlightPlan
				SafeflyFlightPlan sfPlan = (SafeflyFlightPlan) readData;
				// do the planning and return waypoints
				content = sfPlan.getFlightPlan();
				if(content == null){
					return null;
				}
				break;
			
			case CONTENT_GENERATED_PLANS:
				// get SafeflyGeneratedPlans
				SafeflyGeneratedPlans gPlans = (SafeflyGeneratedPlans) readData;
				// do the planning and extract GeneratedPlans
				content = gPlans.getGeneratedPlans();
				if(content == null){
					return null;
				}
				break;
			
			default:
				content = readData;
				break;
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
		
		Message m = Message.getInstance();
		m.setPerformative(performative);
		m.setProtocol(protocol);
		m.setReason(reason);
		m.setConversationID(conversationId);
		m.setContent(content);
		return m;
	}
	
	
}
