package atc.safefly.communication.common.packet;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.LinkedList;
import java.util.List;

import atc.safefly.communication.utils.ApplicationData;
import atc.safefly.utils.SafeflyConstants.Cmd_ID;
import atc.safefly.utils.SafeflyConstants.Cmd_Type;

public class ApplicationPacket extends Packet {
	/*************** Packet Factory registration ************/
	public static final short ID = 0x0001;
	static{
		PacketFactory.registerPacket(new ApplicationPacket());
		
	}
	@Override
	public Packet getInstance() {
		return new ApplicationPacket();
	}
	
	@Override
	public short getApplicationID() {
		return ID;
	}
	
	public ApplicationPacket() {
	}
	/*********************************************************/

	/** Application Data Head - command type and command id, unsigned char */
	public Cmd_Type cmd_type;
	public Cmd_ID cmd_id;
	
	public static final int HEADER_SIZE = 4; 

	/** application data */
	private List<ApplicationData> appData = new LinkedList<ApplicationData>();
	
	public ApplicationPacket(ByteBuffer bb, int length) {
		deserialize(bb, length);
	}

	/**
	 * Creates application packet of given type
	 * 
	 * @param commandId
	 *            - specifies the id of the new command
	 * @param isReply
	 *            - specifies the type
	 * @param params
	 *            - couple of params to be filled in the {@link ApplicationData}
	 *            list
	 */
	public ApplicationPacket(Cmd_ID commandId, Cmd_Type cmdType, int... params) {
		this.cmd_type = cmdType;
		this.cmd_id = commandId;
		if (!cmdType.equals(Cmd_Type.CMDTYPE_REPLY)) {
			ApplicationData data;
			switch (commandId) {
			case CMDID_GRP_SET_PARAM:
				data = new CmdGroupSet();
				data.setValues(params);
				appData.add(data);
				break;
			case CMDID_CHNL_ADD:
				data = new CmdAddChannel();
				data.setValues(params);
				appData.add(data);
				break;
			case CMDID_GRP_DEACT:
				data = new CmdDeactGroup();
				data.setValues(params);
				appData.add(data);
				break;
			case CMDID_RX_DATA:
				data = new CmdRXData();
				data.setValues(params);
				appData.add(data);
				// valid parameters check
				if (params[1] * 2 + 2 != params.length) {
					System.err.println("parameters and data length mismatch: channels count: " + params[1] + " recieved channels parameters: "
							+ (params.length - 2) + " num. of channels parameters should be double the channels count.");
					break;
				}
				// appending respective channel data
				for (int i = 2; i <= params[1] * 2; i += 2) {
					data = new CmdRXDataChannel();
					data.setValues(params[i], params[i + 1]);
					appData.add(data);
				}
				break;
			default:
				throw new RuntimeException("Unknown application packet id: " + commandId);
			}
		} else {
			// reply
			ApplicationData data = new CmdReply();
			data.setValues(params);

			appData.add(data);
		}
	}

	/** Assembly packet to be sent. */
	@Override
	public byte[] serialize() {
		// 4(head)+8(maximal size of ApplicationData chunk = short + 2*byte +
		// int)
		ByteBuffer bb = ByteBuffer.allocate(HEADER_SIZE + appData.size() * 8);
		bb.order(ByteOrder.LITTLE_ENDIAN);
		bb.put(cmd_type.getValue());
		bb.put(cmd_id.getValue());
		bb.put((byte) 0);
		bb.put((byte) 0);

		for (ApplicationData data : appData) {
			bb.put(data.getBytes());
		}

		byte[] pkt = new byte[bb.position()];
		bb.position(0);
		bb.get(pkt);
		return pkt;
	}

	/**
	 * Decode packet bytes into the ApplicationPacket data structure
	 * 
	 * @param array
	 *            - packet bytes
	 * @param from
	 *            - first application data byte index (correspond to command
	 *            type)
	 * @param to
	 *            - last application data byte index
	 */
	public boolean deserialize(ByteBuffer bb, int length) {
		if(length < HEADER_SIZE)
			return false;
		// Command type
		byte cmdType = bb.get();
		// Command ID
		byte cmdId = bb.get();

		// Resolve data type
		ApplicationData data = null;

		switch (cmdId) {
		case 1:
			data = new CmdGroupSet();
			this.cmd_id = Cmd_ID.CMDID_GRP_SET_PARAM;
			break;

		case 2:
			data = new CmdDeactGroup();
			this.cmd_id = Cmd_ID.CMDID_GRP_DEACT;
			break;

		case 3:
			data = new CmdAddChannel();
			this.cmd_id = Cmd_ID.CMDID_CHNL_ADD;
			break;

		case 4:
			data = new CmdRXData();
			this.cmd_id = Cmd_ID.CMDID_RX_DATA;
			break;

		default:
			return false;
		}

		if (cmdType == 0) {
			this.cmd_type = Cmd_Type.CMDTYPE_REQ;
		} else {
			this.cmd_type = Cmd_Type.CMDTYPE_REPLY;
			data = new CmdReply();
		}

		bb.get();
		bb.get(); // skip 2 alignment bytes ~ res2bytes

		length -= HEADER_SIZE;

		int curPos = bb.position();
		data.parseFromBytes(bb, length); // fill
		length -= (bb.position() - curPos);
		this.appData.add(data);
		
		// Read RR channel data
		if (data instanceof CmdRXData) {
			CmdRXData rxd = (CmdRXData)data;
			for (int i = 0; i < rxd.getChannelCount(); i++) {
				curPos = bb.position();
				CmdRXDataChannel rxdch = new CmdRXDataChannel();
				rxdch.parseFromBytes(bb, length); // fill
				length -= (bb.position() - curPos);
				this.appData.add(rxdch);
			}
		}
		return true;
	}

	public Cmd_ID getCmd_id() {
		return cmd_id;
	}

	public Cmd_Type getCmdType() {
		return cmd_type;
	}

	/**
	 * @return List of {@link ApplicationData}
	 */
	public List<ApplicationData> getAppData() {
		return appData;
	}

	/**
	 * Get String representation of the application packet
	 */
	@Override
	public String toString() {
		if (cmd_type.equals(Cmd_Type.CMDTYPE_REPLY)) {
			return "Reply to command: " + cmd_id;
		}
		switch (cmd_id) {
		case CMDID_GRP_SET_PARAM:
			if (appData.get(0) instanceof CmdGroupSet) {
				CmdGroupSet data = (CmdGroupSet) appData.get(0);
				return "GrpId: " + data.grp_id + " refresh: " + data.refresh_interval;
			}
		case CMDID_CHNL_ADD:
			if (appData.get(0) instanceof CmdAddChannel) {
				CmdAddChannel data = (CmdAddChannel) appData.get(0);
				return "GrpId: " + data.grp_id + " channel: " + data.chnl_id;
			}
		case CMDID_GRP_DEACT:
			if (appData.get(0) instanceof CmdDeactGroup) {
				CmdDeactGroup data = (CmdDeactGroup) appData.get(0);
				return "GrpId: " + data.grp_id + " flags: " + data.flags;
			}
		case CMDID_RX_DATA:
			if (appData.get(0) instanceof CmdRXData) {
				CmdRXData data = (CmdRXData) appData.get(0);
				StringBuilder sb = new StringBuilder("GrpId: " + data.grp_id + " channel count: " + data.chnl_cnt);
				// add individual channels
				for (int i = 1; i < appData.size(); i++) {
					CmdRXDataChannel dataChannel = (CmdRXDataChannel) appData.get(i);
					sb.append("\nchannel: " + dataChannel.chnl_id + " value: " + dataChannel.chnl_value);
				}
				return sb.toString();
			}
		default:
			// unknown application packet
			return "Unknown application packet: " + cmd_id;
		}
	}

	// //////////////////////////////////////////////////////////////////////////
	// --------- Definition of different application layer data ----------//
	// //////////////////////////////////////////////////////////////////////////

	/**
	 * Activate group command
	 */
	public class CmdGroupSet extends ApplicationData {
		/** group id 0-7, unsigned linux char ~ java byte */
		byte grp_id;

		/**
		 * Refresh interval of the group (10-65535) [ms x 10] (100ms - 655.35
		 * sec) = 10:55 mm:ss unsigned short
		 */
		int refresh_interval;

		@Override
		public byte[] getBytes() {
			ByteBuffer bb = ByteBuffer.allocate(4);
			bb.order(ByteOrder.LITTLE_ENDIAN);
			bb.put(grp_id);
			bb.put((byte) 0);
			bb.putShort((short) refresh_interval);
			return bb.array();
		}

		@Override
		public void parseFromBytes(ByteBuffer bb, int length) {
			if (length < 4) {
				System.out.println(getClass().getName() + "> Failed to parse bytes!");
				return;
			}
			grp_id = bb.get();
			if (bb.get() != 0) {
				System.out.println(getClass().getName() + "> Failed to parse bytes!");
			}
			refresh_interval = bb.getShort() & 0xFFFF;
		}

		@Override
		public void setValues(int... values) {
			if (values.length > 1) {
				grp_id = (byte) values[0];
				refresh_interval = (short) values[1];
			}
		}

		@Override
		public int getGroupOrChannelId() {
			return grp_id;
		}

		public int getRefreshInt() {
			return refresh_interval;
		}
	}

	/**
	 * Deactivate group command
	 */
	public class CmdDeactGroup extends ApplicationData {
		/** group id 0-7, unsigned char */
		byte grp_id;

		/**
		 * flags bit0 = 1 (clear group), bit0 = 0 (deactivate group, the
		 * channels remain in the group)
		 */
		byte flags;

		@Override
		public byte[] getBytes() {
			return new byte[] { grp_id, flags, 0, 0 };
		}

		@Override
		public void parseFromBytes(ByteBuffer bb, int length) {
			if (length < 4) {
				System.out.println(getClass().getName() + "> Failed to parse bytes!");
				return;
			}
			grp_id = bb.get();
			flags = bb.get();
			if ((bb.get() + bb.get()) != 0) {
				System.out.println(getClass().getName() + "> Failed to parse bytes!");
			}
		}

		@Override
		public void setValues(int... values) {
			if (values.length > 1) {
				grp_id = (byte) values[0];
				flags = (byte) values[1];
			}
		}

		@Override
		public int getGroupOrChannelId() {
			return grp_id;
		}

		public short getFlags() {
			return flags;
		}
	}

	/**
	 * Add channel to a group
	 */
	public class CmdAddChannel extends ApplicationData {
		/** group id 0-7, unsigned char */
		byte grp_id;
		/** channel id to add, unsigned short */
		int chnl_id;

		@Override
		public byte[] getBytes() {
			ByteBuffer bb = ByteBuffer.allocate(4);
			bb.order(ByteOrder.LITTLE_ENDIAN);
			bb.put(grp_id);
			bb.put((byte) 0);
			bb.putShort((short) chnl_id);
			return bb.array();
		}

		@Override
		public void parseFromBytes(ByteBuffer bb, int length) {
			if (length < 4) {
				System.out.println(getClass().getName() + "> Failed to parse bytes!");
				return;
			}
			grp_id = bb.get();
			if (bb.get() != 0) {
				System.out.println(getClass().getName() + "> Failed to parse bytes!");
			}
			chnl_id = bb.getShort() & 0xFFFF;
		}

		@Override
		public void setValues(int... values) {
			if (values.length > 1) {
				grp_id = (byte) values[0];
				chnl_id = (short) values[1];
			}
		}

		@Override
		public int getGroupOrChannelId() {
			return grp_id;
		}

		public int getChnlId() {
			return chnl_id;
		}
	}

	/**
	 * Head of RX measured data
	 */
	class CmdRXData extends ApplicationData {
		/** group id 0-7, unsigned char */
		byte grp_id;
		/** channel count in the group, unsigned char */
		byte chnl_cnt;

		@Override
		public byte[] getBytes() {
			return new byte[] { grp_id, chnl_cnt, 0, 0 };
		}

		@Override
		public void parseFromBytes(ByteBuffer bb, int length) {
			if (length < 4) {
				System.out.println(getClass().getName() + "> Failed to parse bytes!");
				return;
			}
			grp_id = bb.get();
			chnl_cnt = bb.get();
			if ((bb.get() + bb.get()) != 0) {
				System.out.println(getClass().getName() + "> Failed to parse bytes!");
			}
		}

		@Override
		public void setValues(int... values) {
			if (values.length > 1) {
				grp_id = (byte) values[0];
				chnl_cnt = (byte) values[1];
			}
		}

		@Override
		public int getGroupOrChannelId() {
			return grp_id;
		}
		
		public int getChannelCount() {
			return chnl_cnt & 0xFF;
		}
	}

	/**
	 * RX measured data
	 */
	public class CmdRXDataChannel extends ApplicationData {
		/** received channel id, unsigned char */
		short chnl_id;
		/** measurement value, int */
		int chnl_value;

		@Override
		public byte[] getBytes() {
			ByteBuffer bb = ByteBuffer.allocate(8);
			bb.order(ByteOrder.LITTLE_ENDIAN);
			bb.putShort(chnl_id);
			bb.put((byte) 0);
			bb.put((byte) 0);
			bb.putInt(chnl_value);
			return bb.array();
		}

		@Override
		public void parseFromBytes(ByteBuffer bb, int length) {
			if (length < 8) {
				System.out.println(getClass().getName() + "> Failed to parse bytes!");
				return;
			}
			chnl_id = bb.getShort();
			if ((bb.get() + bb.get()) != 0) {
				System.out.println(getClass().getName() + "> Failed to parse bytes!");
			}
			chnl_value = bb.getInt();
		}

		@Override
		public void setValues(int... values) {
			if (values.length > 1) {
				chnl_id = (short) values[0];
				chnl_value = values[1];
			}
		}

		@Override
		public int getGroupOrChannelId() {
			return chnl_id;
		}

		public int getValue() {
			return chnl_value;
		}
	}

	/**
	 * Reply
	 */
	class CmdReply extends ApplicationData {
		byte flags; // currently unused

		@Override
		public byte[] getBytes() {
			return new byte[] { flags };
		}

		@Override
		public void parseFromBytes(ByteBuffer bb, int length) {
			if (length < 1) {
				System.out.println(getClass().getName() + "> Failed to parse bytes!");
				return;
			}
			flags = bb.get();
		}

		@Override
		public void setValues(int... values) {
			if (values.length > 0) {
				flags = (byte) values[0];
			}
		}

		public int getGroupOrChannelId() {
			return -1;
		}
	}
	
	@Override
	public int getSize() {
		throw new UnsupportedOperationException();
	}
	
	
	@Override
	public int getHeaderSize() {
		return HEADER_SIZE;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = super.hashCode();
		result = prime * result + ((appData == null) ? 0 : appData.hashCode());
		result = prime * result + ((cmd_id == null) ? 0 : cmd_id.hashCode());
		result = prime * result + ((cmd_type == null) ? 0 : cmd_type.hashCode());
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (!super.equals(obj))
			return false;
		if (getClass() != obj.getClass())
			return false;
		ApplicationPacket other = (ApplicationPacket) obj;
		if (appData == null) {
			if (other.appData != null)
				return false;
		} else if (!appData.equals(other.appData))
			return false;
		if (cmd_id != other.cmd_id)
			return false;
		if (cmd_type != other.cmd_type)
			return false;
		return true;
	}
	
	
}
