package atc.safefly.communication.common.packet;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.Arrays;

public class FragmentPacket extends AddressedPacket {

	public static final int HEADER_SIZE = 5;

	/*************** Packet Factory registration ************/
	private static final short ID = 0x0101;
	static {
		PacketFactory.registerPacket(new FragmentPacket());
	}

	public FragmentPacket() {
	}

	@Override
	public Packet getInstance() {
		return new FragmentPacket();
	}

	@Override
	public short getApplicationID() {
		return ID;
	}
	
	public FragmentPacket(short dstAddr, short srcAddr, byte[] data){
		this.srcAddr = srcAddr;
		this.dstAddr = dstAddr;
		this.data = data;
	}
	

	/*********************************************************/

	/**
	 * Acknowledge flag for acknowledgement indication
	 */
	public static final byte ACK = 0x01;
	/**
	 * Push flag for message end indication
	 */
	public static final byte PSH = 0x02;
	/**
	 * Reset flag for indicate discard receiver buffer without deserialization
	 * of last message.
	 */
	public static final byte RST = 0x04;
	/**
	 * Synchronisation flag for handshake
	 */
	public static final byte SYN = 0x08;
	/**
	 * Finish flag for closing relation
	 */
	public static final byte NOACK = 0x10;
	/**
	 * Data ignore flag for maintenance packets only
	 */
	public static final byte DIG = 0x20;

	private short sequenceNr;
	private short ackNr;

	/**
	 * Fragmentation flags bytes with set/get flag methods
	 */
	protected byte flags;

	public boolean getFlag(byte flag) {
		return ((flags & flag) != 0);
	}

	public void setFlag(byte flag) {
		flags |= flag;
	}

	public void clearFlag(byte flag) {
		flags &= ~flag;
	}

	public int getSequenceNr() {
		return sequenceNr & 0xFFFF;
	}

	public void setSequenceNr(short sequenceNr) {
		this.sequenceNr = sequenceNr;
	}

	public int getAckNr() {
		return ackNr & 0xFFFF;
	}

	public void setAckNr(short ackNr) {
		this.ackNr = ackNr;
	}

	@Override
	public byte[] serialize() {
		int dLength = 0;
		if (data != null) {
			dLength = data.length;
		}
		ByteBuffer bb = ByteBuffer.allocate(9 + dLength);
		bb.putShort(dstAddr);
		bb.putShort(srcAddr);
		bb.putShort(sequenceNr);
		bb.putShort(ackNr);
		bb.put(flags);
		if (data != null) {
			bb.put(data);
		}

		return bb.array();
	}

	@Override
	public boolean deserialize(ByteBuffer bb, int length) {
		if(length < getHeaderSize())
			return false;
		bb.order(ByteOrder.BIG_ENDIAN);
		dstAddr = bb.getShort();
		srcAddr = bb.getShort();
		sequenceNr = bb.getShort();
		ackNr = bb.getShort();
		flags = bb.get();
		data = new byte[length - 9];
		bb.get(data);
		return true;
	}

	public FragmentPacket clone() {
		FragmentPacket clone = new FragmentPacket();
		clone.sequenceNr = sequenceNr;
		clone.ackNr = ackNr;
		clone.flags = flags;

		clone.srcAddr = srcAddr;
		clone.dstAddr = dstAddr;
		clone.data = data;

		return clone;
	}

	public FragmentPacket deepClone() {
		FragmentPacket clone = new FragmentPacket();
		clone.sequenceNr = sequenceNr;
		clone.ackNr = ackNr;
		clone.flags = flags;

		clone.srcAddr = srcAddr;
		clone.dstAddr = dstAddr;
		if (data != null) {
			clone.data = Arrays.copyOf(data, data.length);
		}

		return clone;
	}

	@Override
	public int getSize() {
		return super.getSize() + HEADER_SIZE;
	}

	@Override
	public int getHeaderSize() {
		return super.getHeaderSize() + HEADER_SIZE;
	}

	@Override
	public String toString() {
		String p = String.format("Fragment: [SeqNr.: %d] [AckNr.: %d]", sequenceNr & 0xFFFF, ackNr & 0xFFFF);
		if ((flags & ACK) > 0) {
			p += "[ACK]";
		}
		if ((flags & PSH) > 0) {
			p += "[PSH]";
		}
		if ((flags & RST) > 0) {
			p += "[RST]";
		}
		if ((flags & SYN) > 0) {
			p += "[SYN]";
		}
		if ((flags & NOACK) > 0) {
			p += "[NOACK]";
		}
		if ((flags & DIG) > 0) {
			p += "[DIG]";
		}

		return String.format("%s {%s}", p, super.toString());

	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = super.hashCode();
		result = prime * result + sequenceNr;
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (!super.equals(obj))
			return false;
		if (getClass() != obj.getClass())
			return false;
		FragmentPacket other = (FragmentPacket) obj;
		if (sequenceNr != other.sequenceNr)
			return false;
		return true;
	}

}
