package atc.safefly.communication.common.packet;

import java.nio.ByteBuffer;
import java.util.BitSet;
import java.util.Random;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.PriorityBlockingQueue;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.Condition;
import java.util.concurrent.locks.ReentrantLock;

import org.apache.log4j.Logger;

import atc.safefly.communication.utils.MessageRecord;

public class MessageConnector extends Connector {
	private static enum State {
		LISTEN, SYN_SENT, SYN_RECEIVED, ESTABLISHED
	}

	/********************** CONSTANTS *******************************/
	private static final int MAX_MESSAGE_SIZE = 32768;// Bytes
	private static final int INIT_TIMEOUT = 10000;// ms
	private static final int WIN_SIZE = 5;
	private static final int INTERPACKET_WAIT = 5;// ms
	private static final int REPEAT_TIMEOUT = 400;// ms
	private static final long MAINTENANCE_PACKET_TIMEOUT = 50;// ms
	/***************************************************************/

	private static final Logger LOGGER = Logger.getLogger(MessageConnector.class);
	/**
	 * Maximal sequence number unsigned short value from {@link FragmentPacket}
	 */
	private static final int MAX_SEQ_NR = 0xFFFF;

	/******************* CONNECTION DESCRIPTORS ********************/
	protected PacketConnector subConnector;
	protected MessageConnectionOwner owner;

	protected short sourceAdr;
	protected short destinationAdr;

	protected final int maxPacketSize = 255 - 19;
	/***************************************************************/

	/********************** IN/OUT MESSAGE RECORDS QUEUES **********/
	protected PriorityBlockingQueue<MessageRecord> sendMessageQueue;
	protected LinkedBlockingQueue<MessageRecord> receiveMessageQueue;
	/***************************************************************/

	/********************** OUTGOING STRUCTURES ********************/
	protected MarkedBlockingList<FragmentPacket> window;
	protected volatile int sequenceNr;
	/***************************************************************/

	/******************** INCOMMING STRUCTURES *********************/
	protected BitSet receivedSeq;
	protected FragmentPacket[] receivedFragments;
	protected volatile int ackNr;
	protected ByteBuffer receiveBuffer = ByteBuffer.allocateDirect(MAX_MESSAGE_SIZE);
	/***************************************************************/

	/******************** EXECUTING THREADS ************************/
	protected Thread sendingThread;
	protected Thread flushingThread;
	protected Thread pushingThread;
	protected TimedTask maintenanceTask;
	/**************************************************************/

	FragmentPacket maintenancePacket = new FragmentPacket();

	/******************** SYSTEM STATE VARIABLES ******************/
	protected State state;
	protected Object stateLock = new Object();

	/**************************************************************/

	public MessageConnector(PacketConnector subConnector, MessageConnectionOwner owner, short sourceAdr, short destinationAdr) {
		this.subConnector = subConnector;
		// this.maxPacketSize = subConnector.getMaxPacketSize();
		this.owner = owner;

		this.sourceAdr = sourceAdr;
		this.destinationAdr = destinationAdr;
		setState(State.LISTEN);

		sequenceNr = new Random().nextInt(MAX_SEQ_NR);

		maintenancePacket.flags = FragmentPacket.ACK | FragmentPacket.NOACK;
		maintenancePacket.srcAddr = sourceAdr;
		maintenancePacket.dstAddr = destinationAdr;
	}

	@Override
	public synchronized boolean connect() {
		if (!isConnected()) {
			if (subConnector.connect() && open()) {
				isConnected = true;
			} else {
				dispose();
			}
		}
		return isConnected;
	}

	@Override
	public synchronized void dispose() {
		if (isConnected()) {
			close();
		}
		subConnector.dispose();
		isConnected = false;
	}

	@Override
	public boolean isConnected() {
		if (!subConnector.isConnected() && isConnected) {
			close();
		}
		return isConnected;
	}

	@Override
	public synchronized void close() {
		isConnected = false;
		// Clear incoming structures
		if (receivedSeq != null) {
			receivedSeq.clear();
		}
		receivedFragments = null;
		ackNr = -1;
		setState(State.LISTEN);
		try {
			if(maintenanceTask != null){
				maintenanceTask.interrupt();
			}
			if(flushingThread != null){
				flushingThread.interrupt();
				flushingThread.join();
			}
			if(sendingThread != null){
				sendingThread.interrupt();
				sendingThread.join();
			}
			if(pushingThread != null){
				pushingThread.interrupt();
				pushingThread.join();
			}
		} catch (InterruptedException e) {
			e.printStackTrace();
		}
		if (window != null) {
			window.clear();
			window = null;
		}
		while (!sendMessageQueue.isEmpty()) {
			MessageRecord mr;
			try {
				mr = sendMessageQueue.take();
				mr.setState(MessageRecord.State.FAILED);
			} catch (InterruptedException e) {
			}
		}
		if (receiveMessageQueue != null)
			receiveMessageQueue.clear();
		if (sendMessageQueue != null)
			sendMessageQueue.clear();
	}

	@Override
	protected boolean open() {
		sendMessageQueue = new PriorityBlockingQueue<MessageRecord>();
		receiveMessageQueue = new LinkedBlockingQueue<MessageRecord>();
		ackNr = -1;
		window = new MarkedBlockingList<FragmentPacket>(WIN_SIZE);
		flushingThread = new Thread(new Runnable() {

			@Override
			public void run() {
				try {
					flush();
				} catch (InterruptedException e) {
					// System.err.println(Thread.currentThread().getName() +
					// " is interrupted!");
				}
			}
		}, getClass().getSimpleName() + ": Flushing thread");
		flushingThread.start();

		pushingThread = new Thread(new Runnable() {

			@Override
			public void run() {
				pushing();
			}
		}, getClass().getSimpleName() + ": Pushing thread");
		pushingThread.start();

		sendingThread = new Thread(new Runnable() {

			@Override
			public void run() {
				try {
					enqueue();
				} catch (InterruptedException e) {
					// System.err.println(Thread.currentThread().getName() +
					// " is interrupted!");
				}
			}
		}, getClass().getSimpleName() + ": Sending thread");
		sendingThread.start();

		maintenanceTask = new TimedTask(new Runnable() {

			@Override
			public void run() {
				sendMaintenancePacket();
			}
		}, MAINTENANCE_PACKET_TIMEOUT);
		receivedFragments = new FragmentPacket[MAX_SEQ_NR];
		receivedSeq = new BitSet(MAX_SEQ_NR);

		isConnected = true;
		return true;
	}

	public boolean sendMessage(MessageRecord mr) {
		if (!isConnected()) {
			LOGGER.warn("not connected - cannot send message");
			return false;
		}
		return sendMessageQueue.add(mr);
	}

	protected void enqueue() throws InterruptedException {
		while (!Thread.interrupted()) {
			// Get first message from queue
			MessageRecord currentSendingMessage = sendMessageQueue.take();

			// Wait for connection establishing when it is necessary
			synchronized (stateLock) {
				if (state != State.ESTABLISHED) {
					// Wait for handshake
					if (!initConnection()) {
						System.err.println(getClass().getName() + ": Failed to initialize connection to remote host!");
						currentSendingMessage.setState(MessageRecord.State.FAILED);
						continue;
					}
				}
			}

			// Remove message timed out
			if (currentSendingMessage.isExpired()) {
				currentSendingMessage.setState(MessageRecord.State.EXPIRED);
				currentSendingMessage = null;
				continue;
			}

			currentSendingMessage.setState(MessageRecord.State.SENDING);

			// Split message to fragments and push to flushing thread
			byte[] sPacket = currentSendingMessage.getPacket().serialize();
			ByteBuffer bb = ByteBuffer.allocate(sPacket.length + 2);
			bb.putShort(currentSendingMessage.getPacket().getApplicationID());
			bb.put(sPacket);
			bb.position(0);

			int restBytes = bb.limit() - bb.position();
			long timeOut = currentSendingMessage.getTimeout();
			while (restBytes > 0 && timeOut > 0 && !Thread.interrupted()) {
				// Compute fragment size
				int fragmentSize = maxPacketSize;
				FragmentPacket fragment = new FragmentPacket();
				if (restBytes <= maxPacketSize) {
					fragmentSize = restBytes;
					fragment.flags |= FragmentPacket.PSH;
				}
				fragment.flags |= FragmentPacket.ACK;
				byte[] fragmentData = new byte[fragmentSize];
				bb.get(fragmentData);
				fragment.data = fragmentData;

				timeOut = enqueueFragment(fragment, timeOut);
				restBytes = bb.limit() - bb.position();
			}

			if (restBytes > 0) {
				// Sending failed
				currentSendingMessage.setState(MessageRecord.State.FAILED);
			} else {
				currentSendingMessage.setState(MessageRecord.State.SENT);
			}
		}
	}

	private boolean initConnection() throws InterruptedException {
		synchronized (stateLock) {
			if (state == State.LISTEN) {
				// Start synchronization
				FragmentPacket syn = new FragmentPacket();
				syn.setFlag(FragmentPacket.SYN);
				long t = enqueueFragment(syn, INIT_TIMEOUT);
				if (t <= 0L) {
					return false;
				}
				setState(State.SYN_SENT);
				// Wait for process completion
				stateLock.wait(INIT_TIMEOUT);

			} else if (state == State.SYN_RECEIVED) {
				// Wait for process completion
				stateLock.wait(INIT_TIMEOUT);
			} else {
				System.err.println(getClass().getName() + ": Unexpected state for connection initialization: " + state.toString());
			}

			return (state == State.ESTABLISHED);
		}
	}

	private int enqueueFragment(FragmentPacket fragment) throws InterruptedException {
		// Increment sequence number
		window.lock();
		try {
			sequenceNr = (sequenceNr + 1) % MAX_SEQ_NR;
			fragment.setSequenceNr((short) sequenceNr);
			// Add routing information
			fragment.srcAddr = sourceAdr;
			fragment.dstAddr = destinationAdr;

			window.addLast(fragment);
			return sequenceNr;
		} finally {
			window.unlock();
		}
	}

	private long enqueueFragment(FragmentPacket fragment, long timeout) throws InterruptedException {
		// Increment sequence number
		window.lock();
		try {
			sequenceNr = (sequenceNr + 1) % MAX_SEQ_NR;
			fragment.setSequenceNr((short) sequenceNr);
			// Add routing information
			fragment.srcAddr = sourceAdr;
			fragment.dstAddr = destinationAdr;

			timeout = window.addLast(fragment, timeout);

			if (timeout <= 0L) {
				// sending failed, reset receiving on distant site
				fragment.data = new byte[0];
				fragment.setFlag(FragmentPacket.RST);
				window.addLast(fragment);
			}
			return timeout;
		} finally {
			window.unlock();
		}
	}

	private void flush() throws InterruptedException {
		while (!Thread.interrupted()) {
			FragmentPacket fragment = window.getNext(REPEAT_TIMEOUT);
			if (fragment != null) {
				fragment.setAckNr((short) ackNr);
				maintenanceTask.stop();
				subConnector.sendPacket(fragment);
				Thread.sleep(INTERPACKET_WAIT);
			} else {
				window.reset();
			}
		}
	}

	protected void sendMaintenancePacket() {
		maintenancePacket.setAckNr((short) ackNr);
		subConnector.sendPacket(maintenancePacket);
	}

	public synchronized void fragmentReceived(FragmentPacket packet) {
		if (!isConnected) {
			return;
		}

		/********************* UPDATE RECEIVED ACKs ************************/
		if (packet.getFlag(FragmentPacket.ACK)) {

			// Sequential acknowledgement
			int ack = packet.getAckNr();
			window.lock();
			try {
				FragmentPacket fragment = window.getFirst();
				while (fragment != null) {
					int dist = (ack - fragment.getSequenceNr() + MAX_SEQ_NR) % MAX_SEQ_NR;
					if (dist <= WIN_SIZE) {
						window.removeFirst();
						fragment = window.getFirst();
					} else {
						break;
					}
				}
			} finally {
				window.unlock();
			}
		}
		/*******************************************************************/
		if (!packet.getFlag(FragmentPacket.NOACK)) {
			maintenanceTask.start();
			/********************** UPDATE RECEIVED FRAGMENTS ******************/
			int seqNr = packet.getSequenceNr();
			if (ackNr < 0) {
				// First packet
				ackNr = ((seqNr - 1) + MAX_SEQ_NR) % MAX_SEQ_NR;
			}
			int dist = ((seqNr - ackNr - 1) + MAX_SEQ_NR) % MAX_SEQ_NR;
			if (dist <= WIN_SIZE) {
				// Currently not received packet
				receivedFragments[seqNr] = packet;

				// Check received sequence
				receivedSeq.set(seqNr);
				int nextAckNr = (ackNr + 1) % MAX_SEQ_NR;
				while (receivedSeq.get(nextAckNr)) {
					// shift ackNr to last received packet
					receivedSeq.clear(ackNr);
					ackNr = nextAckNr;
					nextAckNr = (ackNr + 1) % MAX_SEQ_NR;
					// Check PUSH flag for currently received sequence
					if (receivedFragments[ackNr].getFlag(FragmentPacket.PSH)) {
						// Complete fragment sequence is ready, compose
						// packet
						composeFragments();
					} else if (receivedFragments[ackNr].getFlag(FragmentPacket.RST)) {
						int ptr = ackNr;
						// Decrease in module while start is found
						while (receivedFragments[ptr] != null) {
							receivedFragments[ptr] = null;
							ptr = ((ptr - 1) + MAX_SEQ_NR) % MAX_SEQ_NR;
						}
					}
				}
			} else if (receivedFragments[seqNr] != null && receivedFragments[seqNr].flags == packet.flags) {
				// Do not process flags aggain
				return;
			}
			/*******************************************************************/
			try {
				processFlags(packet);
			} catch (InterruptedException e) {
				return;
			}
		}
	}

	private void composeFragments() {
		receiveBuffer.clear();
		int ptr = ackNr;
		// Decrease in module while start is found
		while (receivedFragments[ptr] != null) {
			ptr = ((ptr - 1) + MAX_SEQ_NR) % MAX_SEQ_NR;
		}
		// Increase to first not null field
		ptr++;
		ptr %= MAX_SEQ_NR;
		// Read data into buffer
		while (ptr != ackNr) {
			if (!receivedFragments[ptr].getFlag(FragmentPacket.DIG)) {
				receiveBuffer.put(receivedFragments[ptr].getData());
			}
			receivedFragments[ptr] = null;
			ptr++;
			ptr %= MAX_SEQ_NR;
		}
		// Read last data fragment
		if (!receivedFragments[ptr].getFlag(FragmentPacket.DIG)) {
			receiveBuffer.put(receivedFragments[ptr].getData());
		}
		receivedFragments[ptr] = null;

		// Extract packet
		int packetLength = receiveBuffer.position() - 2;// dataLength + 2B
														// packet ID
		receiveBuffer.position(0);
		short packetID = receiveBuffer.getShort();
		try {
			PacketInterface packet = PacketFactory.getPacket(packetID);

			if (packet == null) {
				return;
			}
			packet.deserialize(receiveBuffer, packetLength);
			receiveMessageQueue.put(new MessageRecord(packet, sourceAdr, destinationAdr, 0));
		} catch (InterruptedException e) {
			return;
		} catch (UnsupportedOperationException e) {
			e.printStackTrace();
			return;
		}
	}

	private void processFlags(FragmentPacket packet) throws InterruptedException {
		synchronized (stateLock) {
			switch (state) {
				case LISTEN:
					if (packet.getFlag(FragmentPacket.SYN) && !packet.getFlag(FragmentPacket.ACK)) {
						// SYN packet received
						setState(State.SYN_RECEIVED);
						FragmentPacket ack = new FragmentPacket();
						ack.setFlag(FragmentPacket.ACK);
						ack.setFlag(FragmentPacket.SYN);
						enqueueFragment(ack);
					} else {
						if (window.isEmpty()) {
							// Send dummy message to  invoke handshake
							sendMessage(new MessageRecord(new Packet(), destinationAdr, sourceAdr, 0));
						}
					}
					break;
				case SYN_SENT:
					if (packet.getFlag(FragmentPacket.SYN)) {
						if (packet.getFlag(FragmentPacket.ACK)) {
							// Connection established
							FragmentPacket ack = new FragmentPacket();
							ack.setFlag(FragmentPacket.ACK);
							enqueueFragment(ack);
							setState(State.ESTABLISHED);
							stateLock.notifyAll();
						} else {
							// Connection establishing from remote host
							// Check master in accordance to sequenceNr
							if (sequenceNr < packet.getSequenceNr()) {
								// Slave mode - do the same as in LISTEN case
								setState(State.SYN_RECEIVED);
								window.clear();
								FragmentPacket ack = new FragmentPacket();
								ack.setFlag(FragmentPacket.ACK);
								ack.setFlag(FragmentPacket.SYN);
								enqueueFragment(ack);
							} else if (sequenceNr == packet.getSequenceNr()) {
								// Deal the tie by new random generated SequenceNr
								sequenceNr = new Random().nextInt(MAX_SEQ_NR);
								// Reinitialize connection
								setState(State.LISTEN);
								window.clear();
								try {
									initConnection();
								} catch (InterruptedException e) {
									return;
								}
							}// else Master-do nothing / remote host should send
								// SYN+ACK
						}
					} else {
						System.err.println(getClass().getName() + ": Unsynchronized packet received! in " + state);
					}
					break;
				case SYN_RECEIVED:
					if (packet.getFlag(FragmentPacket.ACK)) {
						setState(State.ESTABLISHED);
						stateLock.notifyAll();
					}
					break;
				case ESTABLISHED:
					if (packet.getFlag(FragmentPacket.SYN) && !packet.getFlag(FragmentPacket.ACK)) {
						System.err.println("RESYNCHRONIZED");
						// RESYNC initialized;
						ackNr = packet.getSequenceNr();
						// Clear incomming queues
						receivedSeq.clear();
						receivedFragments = new FragmentPacket[MAX_SEQ_NR];
						window.clear();
						ackNr = -1;
						setState(State.SYN_RECEIVED);
						FragmentPacket ack = new FragmentPacket();
						ack.setFlag(FragmentPacket.ACK);
						ack.setFlag(FragmentPacket.SYN);
						enqueueFragment(ack);
					}
					break;
			}
		}
	}

	//		public synchronized void fragmentReceived(FragmentPacket packet) {
	//			/********************* UPDATE RECEIVED ACKs ************************/
	//			if (packet.getFlag(FragmentPacket.ACK)) {
	//				// Check ACK
	//				synchronized (unreceivedAck) {
	//					// Accidental acknowledgement
	//					if (packet.getFlag(FragmentPacket.DIG)) {
	//						ByteBuffer bb = ByteBuffer.wrap(packet.getData());
	//						while (bb.hasRemaining()) {
	//							unreceivedAck.clear(bb.getShort() & MAX_SEQ_NR);
	//						}
	//					}
	//					// Sequential acknowledgement
	//					int ack = packet.getAckNr();
	//					// Clear each awaiting ACK smaller than received
	//					while (fragmentsMap[ack] != null) {
	//						synchronized (fragmentsMap[ack]) {
	//							fragmentsMap[ack].notifyAll();
	//							fragmentsMap[ack] = null;
	//							unreceivedAck.clear(ack);
	//						}
	//						// Decrement in module MAX_SEQ_NR
	//						ack = ((ack - 1) + MAX_SEQ_NR) % MAX_SEQ_NR;
	//					}
	//	
	//					unreceivedAck.notifyAll();
	//				}
	//			}
	//			/*******************************************************************/
	//			if (!packet.getFlag(FragmentPacket.NOACK)) {
	//				/********************** UPDATE RECEIVED FRAGMENTS ******************/
	//	
	//				synchronized (receivedSeq) {
	//					int seqNr = packet.getSequenceNr();
	//					if (ackNr < 0) {
	//						// First packet
	//						ackNr = ((packet.getSequenceNr() - 1) + MAX_SEQ_NR) % MAX_SEQ_NR;
	//					}
	//					int dist = ((seqNr - ackNr - 1) + MAX_SEQ_NR) % MAX_SEQ_NR;
	//					if (dist < WIN_SIZE) {
	//						// Currently not received packet
	//						// Not fragmented packet without sequence number
	//						receivedFragments[seqNr] = packet;
	//	
	//						// Check received sequence
	//						receivedSeq.set(seqNr);
	//						int nextAckNr = (ackNr + 1) % MAX_SEQ_NR;
	//						while (receivedSeq.get(nextAckNr)) {
	//							// shift ackNr to last received packet
	//							receivedSeq.clear(ackNr);
	//							ackNr = nextAckNr;
	//							nextAckNr = (ackNr + 1) % MAX_SEQ_NR;
	//							// Check PUSH flag for currently received sequence
	//							if (receivedFragments[ackNr].getFlag(FragmentPacket.PSH)) {
	//								// Complete fragment sequence is ready, compose
	//								// packet
	//								composeFragments();
	//							}
	//						}
	//					} else if (receivedFragments[seqNr] != null && receivedFragments[seqNr].flags == packet.flags) {
	//						// Do not process flags aggain
	//						return;
	//					}
	//				}
	//				/*******************************************************************/
	//			}
	//	
	//			/******************** PROCESS STATE SENSITIVE STRUCTURES ***********/
	//			synchronized (stateLock) {
	//				switch (state) {
	//					case LISTEN:
	//						if (packet.getFlag(FragmentPacket.SYN) && !packet.getFlag(FragmentPacket.ACK)) {
	//							// SYN packet received
	//							setState(State.SYN_RECEIVED);
	//							ackNr = packet.getSequenceNr();
	//							FragmentPacket ack = new FragmentPacket();
	//							ack.setFlag(FragmentPacket.ACK);
	//							ack.setFlag(FragmentPacket.SYN);
	//							synReceivedAck = sendFragment(ack);
	//						} else {
	//							if (currentSendingMessage == null) {
	//								sendMessage(new MessageRecord(new Packet(), destinationAdr, sourceAdr, 0));
	//							}
	//						}
	//						break;
	//					case SYN_SENT:
	//						if (packet.getFlag(FragmentPacket.SYN)) {
	//							if (packet.getFlag(FragmentPacket.ACK)) {
	//								// Connection established
	//								setState(State.ESTABLISHED);
	//								ackNr = packet.getSequenceNr();
	//								FragmentPacket ack = new FragmentPacket();
	//								ack.setFlag(FragmentPacket.ACK);
	//								sendFragment(ack);
	//								stateLock.notifyAll();
	//							} else {
	//								// Connection establishing from remote host
	//								// Check master in accordance to sequenceNr
	//								if (sequenceNr < packet.getSequenceNr()) {
	//									// Slave mode - do the same as in LISTEN case
	//									setState(State.SYN_RECEIVED);
	//									FragmentPacket ack = new FragmentPacket();
	//									ack.setFlag(FragmentPacket.ACK);
	//									ack.setFlag(FragmentPacket.SYN);
	//									sendFragment(ack);
	//								} else if (sequenceNr == packet.getSequenceNr()) {
	//									// Deal the tie by new random generated SequenceNr
	//									sequenceNr = new Random().nextInt(MAX_SEQ_NR);
	//									// Reinitialize connection
	//									setState(State.LISTEN);
	//									try {
	//										initConnection();
	//									} catch (InterruptedException e) {
	//										return;
	//									}
	//								}// else Master-do nothing / remote host should send
	//									// SYN+ACK
	//							}
	//						} else {
	//							System.err.println(getClass().getName() + ": Unsynchronized packet received! in " + state);
	//						}
	//						break;
	//					case SYN_RECEIVED:
	//						synchronized (unreceivedAck) {
	//							if (!unreceivedAck.get(synReceivedAck)) {
	//								setState(State.ESTABLISHED);
	//								stateLock.notifyAll();
	//							}
	//						}
	//						break;
	//					case ESTABLISHED:
	//						if (packet.getFlag(FragmentPacket.SYN) && !packet.getFlag(FragmentPacket.ACK)) {
	//							System.err.println("RESYNCHRONIZED");
	//							// RESYNC initialized;
	//							ackNr = -1;
	//							setState(State.SYN_RECEIVED);
	//							ackNr = packet.getSequenceNr();
	//							// Clear all queues
	//							receivedSeq.clear();
	//							receivedFragments = new FragmentPacket[MAX_SEQ_NR];
	//	
	//							if (fragmentsMap[currentSendingLastSegment] != null) {
	//								synchronized (fragmentsMap[currentSendingLastSegment]) {
	//									sendMessageQueue.add(currentSendingMessage);
	//									Packet p = fragmentsMap[currentSendingLastSegment];
	//	
	//									synchronized (unreceivedAck) {
	//										fragmentsMap = new FragmentPacket[MAX_SEQ_NR];
	//										unreceivedAck.clear();
	//										unreceivedAck.notifyAll();
	//									}
	//									p.notifyAll();
	//								}
	//							}
	//	
	//							FragmentPacket ack = new FragmentPacket();
	//							ack.setFlag(FragmentPacket.ACK);
	//							ack.setFlag(FragmentPacket.SYN);
	//							synReceivedAck = sendFragment(ack);
	//						}
	//						break;
	//				}
	//			}
	//		}

	/*
	protected void sending() throws InterruptedException {
		while (!Thread.interrupted()) {

			// Wait for message
			currentSendingMessage = sendMessageQueue.take();

			synchronized (stateLock) {
				if (state != State.ESTABLISHED) {
					// Wait for handshake
					if (!initConnection()) {
						System.err.println(getClass().getName() + ": Failed to initialize connection to remote host!");
						currentSendingMessage.setState(MessageRecord.State.FAILED);
						continue;
					}
				}
			}

			// Remove messages timed out
			if (currentSendingMessage.isExpired()) {
				currentSendingMessage.setState(MessageRecord.State.EXPIRED);
				currentSendingMessage = null;
				continue;
			}

			currentSendingMessage.setState(MessageRecord.State.SENDING);
			// Connection established, start sending fragments
			byte[] sPacket = currentSendingMessage.getPacket().serialize();
			ByteBuffer bb = ByteBuffer.allocate(sPacket.length + 2);
			bb.putShort(currentSendingMessage.getPacket().getApplicationID());
			bb.put(sPacket);
			bb.position(0);

			FragmentPacket packet = new FragmentPacket();
			packet.flags = FragmentPacket.ACK;

			int maxSize = maxPacketSize - packet.getHeaderSize();
			while ((bb.limit() - bb.position()) > maxSize) {
				FragmentPacket cPacket = packet.clone();
				byte[] data = new byte[maxSize];
				bb.get(data);
				cPacket.data = data;
				sendFragment(cPacket);
			}
			int restSize = bb.limit() - bb.position();
			byte[] data = new byte[restSize];
			bb.get(data);
			packet.data = data;
			packet.setFlag(FragmentPacket.PSH);
			synchronized (packet) {
				currentSendingLastSegment = sendFragment(packet);
				// Wait for acknowledgement of last packet
				packet.wait(currentSendingMessage.getTimeout());
				if (currentSendingMessage == null) {
					// Resynchronization invoked
					continue;
				}
				if (unreceivedAck.get(currentSendingLastSegment)) {
					currentSendingMessage.setState(MessageRecord.State.EXPIRED);
					synchronized (unreceivedAck) {
						// TODO: Reset sending: remove fragments out of current
						// window

					}

				} else {
					currentSendingMessage.setState(MessageRecord.State.SENT);
				}
				currentSendingMessage = null;
			}

		}

	}

	protected void flushing() throws InterruptedException {
		// Last unacknowledged packet
		int last = 0;
		// Current packet to send
		int current = 0;

		long lastSentTime = System.currentTimeMillis();

		while (!Thread.interrupted()) {
			synchronized (unreceivedAck) {
				while (unreceivedAck.isEmpty()) {
					// Wait for new fragment
					unreceivedAck.wait();
					if (unreceivedAck.isEmpty() && (System.currentTimeMillis() - lastSentTime) > MAINTENANCE_PACKET_TIMEOUT) {
						// Send maintenance unacknowledged packet
						FragmentPacket packet = new FragmentPacket();
						packet.flags = FragmentPacket.ACK | FragmentPacket.NOACK;
						packet.srcAddr = sourceAdr;
						packet.dstAddr = destinationAdr;
						synchronized (receivedSeq) {
							packet.setAckNr((short) ackNr);
						}
						subConnector.sendPacket(packet);
						lastSentTime = System.currentTimeMillis();
						continue;
					}
					// TODO: Test if last == 0, try to decrease in module
					last = unreceivedAck.nextSetBit(0);
					current = last;
				}

				// Move last to first unacknowledged packet
				while (!unreceivedAck.get(last)) {
					last = (last + 1) % MAX_SEQ_NR;
				}

				int dist = ((current - last) + MAX_SEQ_NR) % MAX_SEQ_NR;

				if (dist > WIN_SIZE) {
					// Indicate that last pass the current
					current = last;
					dist = 0;
				} else {
					// Shift current to first unacknowledged packet within
					// window
					while (!unreceivedAck.get(current) && dist <= WIN_SIZE) {
						current = (current + 1) % MAX_SEQ_NR;
						dist++;
					}
				}

				if (dist < WIN_SIZE) {
					// SEND CURRENT PACKET
					synchronized (receivedSeq) {
						fragmentsMap[current].setAckNr((short) ackNr);
					}
					subConnector.sendPacket(fragmentsMap[current]);
					current = (current + 1) % MAX_SEQ_NR;
					lastSentTime = System.currentTimeMillis();
				}

				if (dist == WIN_SIZE || dist == 0) {
					// Wait for change or repeat packets
					unreceivedAck.wait(RESEND_WAIT);
					if (unreceivedAck.get(last)) {
						// start repeat
						current = last;
					}
				}
			}

			Thread.sleep(INTERPACKET_WAIT);
		}
	}

	protected synchronized int sendFragment(FragmentPacket packet) {

		sequenceNr = (sequenceNr + 1) % MAX_SEQ_NR;
		packet.setSequenceNr((short) sequenceNr);

		packet.srcAddr = sourceAdr;
		packet.dstAddr = destinationAdr;

		synchronized (unreceivedAck) {
			fragmentsMap[sequenceNr] = packet;
			unreceivedAck.set(sequenceNr);
			unreceivedAck.notifyAll();
		}
		return sequenceNr;
	}
	*/

	protected void pushing() {
		if (owner != null) {
			while (!Thread.interrupted()) {
				try {
					owner.incommingMessage(receiveMessageQueue.take());
				} catch (InterruptedException e) {
					// System.err.println(Thread.currentThread().getName() +
					// " is interrupted!");
					return;
				}
			}
		}
	}

	private void setState(State state) {
		synchronized (stateLock) {
			System.out.println("State: " + this.state + ">" + state);
			LOGGER.info("State: " + this.state + ">" + state);
			this.state = state;
		}
	}

	private class MarkedBlockingList<E> {
		ListEntry first;
		ListEntry last;
		ListEntry current;
		int size;
		int limit;

		private ReentrantLock lock;
		private Condition added;
		private Condition removed;

		public MarkedBlockingList(int limit) {
			this.limit = limit;
			first = null;
			last = null;
			current = null;
			size = 0;
			lock = new ReentrantLock(true);
			added = lock.newCondition();
			removed = lock.newCondition();
		}

		public void clear() {
			lock.lock();
			try {
				first = null;
				last = null;
				current = null;
				size = 0;
				removed.signalAll();
			} finally {
				lock.unlock();
			}
		}

		public E getNext(long timeout) throws InterruptedException {
			long nanos = TimeUnit.MILLISECONDS.toNanos(timeout);
			lock.lock();
			try {
				while (current == null && nanos > 0L)
					nanos = added.awaitNanos(nanos);
				if (current == null) {
					return null;
				}
				E currentElement = current.element;
				current = current.next;
				return currentElement;
			} finally {
				lock.unlock();
			}
		}

		public boolean isEmpty() {
			lock.lock();
			try {
				return size == 0;
			} finally {
				lock.unlock();
			}
		}

		public void reset() {
			lock.lock();
			try {
				current = first;
			} finally {
				lock.unlock();
			}
		}

		public void addLast(E fragment) throws InterruptedException {
			addLast(fragment, 60000);
		}

		public long addLast(E element, long timeout) throws InterruptedException {
			long nanos = TimeUnit.MILLISECONDS.toNanos(timeout);
			lock.lock();
			try {
				while (size == limit && nanos > 0L)
					nanos = removed.awaitNanos(nanos);
				if (nanos <= 0L) {
					return TimeUnit.NANOSECONDS.toMillis(nanos);
				}

				// Add new element as last
				ListEntry le = new ListEntry();
				le.element = element;
				if (size == 0) {
					first = le;
					last = le;
					current = le;
				} else {
					// Chain with last
					last.next = le;
					last = le;
				}
				if (current == null) {
					current = last;
				}
				size++;
				added.signalAll();
				return TimeUnit.NANOSECONDS.toMillis(nanos); // Rest of timeout in miliseconds
			} finally {
				lock.unlock();
			}
		}

		public E getFirst() {
			lock.lock();
			try {
				return first == null ? null : first.element;
			} finally {
				lock.unlock();
			}
		}

		public boolean removeFirst() {
			lock.lock();
			try {
				if (first != null) {
					first = first.next;
					size--;
					removed.signalAll();
					return true;
				}
				return false;
			} finally {
				lock.unlock();
			}
		}

		public void lock() {
			lock.lock();
		}

		public void unlock() {
			lock.unlock();
		}

		private class ListEntry {
			E element = null;
			ListEntry next = null;
		}
	}

	public class TimedTask {
		private final long timeout;
		private final Runnable task;

		private final ReentrantLock lock = new ReentrantLock();
		private final Condition cond = lock.newCondition();

		private Thread starter;
		private boolean interrupted = false;

		public TimedTask(Runnable task, long timeout) {
			this.task = task;
			this.timeout = timeout;
		}

		public void start() {
			lock.lock();
			try {
				if (interrupted)
					return;
				if (starter == null || !starter.isAlive()) {
					starter = new Thread(new Runnable() {

						@Override
						public void run() {
							countdown();
						}
					});
					starter.start();
				} else {
					cond.signalAll();
				}
			} finally {
				lock.unlock();
			}
		}

		public void stop() {
			lock.lock();
			try {
				if (starter != null) {
					starter.interrupt();
				}
			} finally {
				lock.unlock();
			}
		}

		public void interrupt() {
			lock.lock();
			try {
				stop();
				interrupted = true;
			} finally {
				lock.unlock();
			}
		}

		private void countdown() {
			lock.lock();
			try {
				while (cond.await(timeout, TimeUnit.MILLISECONDS))
					;
				task.run();
			} catch (InterruptedException e) {
				return;
			} finally {
				lock.unlock();
			}
		}

	}

}
