package atc.safefly.communication.common.packet;

import java.io.InputStream;
import java.io.OutputStream;
import java.util.concurrent.LinkedBlockingQueue;

import atc.safefly.communication.common.serial.SerialConnector;

public abstract class PacketConnector extends Connector {

	protected LinkedBlockingQueue<PacketInterface> packetQueue;

	protected InputStream in;
	protected OutputStream out;

	protected PacketConnectionOwner owner;
	protected SerialConnector subConnector;

	protected Thread pushingThread;
	protected Thread readThread;

	public PacketConnector(SerialConnector subConnector, PacketConnectionOwner owner) {
		this.subConnector = subConnector;
		this.owner = owner;
	}
	
	public PacketConnector(SerialConnector subConnector) {
		this(subConnector, null);
	}
	

	@Override
	public boolean connect() {
		if (!isConnected()) {
			if (subConnector.connect() && open()) {
				isConnected = true;
			} else {
				dispose();
			}
		}
		return isConnected;
	}

	@Override
	public synchronized void dispose() {
		if (isConnected) {
			close();
		}
		subConnector.dispose();
		isConnected = false;
		System.err.println("Disposing connector...");
	}

	@Override
	public boolean isConnected() {
		if (!subConnector.isConnected() && isConnected) {
			close();
			isConnected = false;
		}
		return isConnected;
	}

	/**
	 * Provide packet sending
	 * 
	 * @param packet
	 *            - data to send
	 * @return true if success, false otherwise
	 */
	public abstract boolean sendPacket(PacketInterface packet);

	protected abstract void readPacket();

	protected void pushPacket() {
			while (!Thread.interrupted()) {
				try {
					PacketInterface p = packetQueue.take();
					if (owner != null) {
					owner.incommingPacket(p);
					}
				} catch (InterruptedException e) {
					return;
				} catch (NullPointerException e) {
					return;
				}
			}
	}

	@Override
	public void close() {
		readThread.interrupt();
		readThread = null;
		pushingThread.interrupt();
		pushingThread = null;
		packetQueue.clear();
		packetQueue = null;
		in = null;
		out = null;
		subConnector.restoreStreams();
		isConnected = false;
	}

	@Override
	protected boolean open() {
		in = subConnector.getInputStream();
		out = subConnector.getOutputStream();
		if (in == null || out == null) {
			System.err.println("Failed to open BasicDataConnector, streams already occupied!");
			return false;
		}
		packetQueue = new LinkedBlockingQueue<PacketInterface>();
		readThread = new Thread(new Runnable() {

			@Override
			public void run() {
				readPacket();
			}
		}, getClass().getSimpleName()+": Reading thread");

		pushingThread = new Thread(new Runnable() {

			@Override
			public void run() {
				pushPacket();
			}
		}, getClass().getSimpleName()+": Pushing thread");

		pushingThread.start();
		readThread.start();
		return true;
	}
	
	public abstract int getMaxPacketSize();
	
	public void setOwner(PacketConnectionOwner owner) {
		this.owner = owner;
	}

}
