package atc.safefly.communication.common.packet;

import java.nio.ByteBuffer;
import java.nio.ByteOrder;
import java.util.Arrays;
import java.util.Random;

public class TraceRoutePacket extends AddressedPacket {

	public static enum Type {
		DIRECT, BUMERANG, CYCLE
	}

	/*************** Packet Factory registration ************/
	private static final short ID = 0x1001;
	static {
		PacketFactory.registerPacket(new TraceRoutePacket());
	}

	public TraceRoutePacket() {
	}

	@Override
	public Packet getInstance() {
		return new TraceRoutePacket();
	}

	@Override
	public short getApplicationID() {
		return ID;
	}

	/*******************************************************/

	private long uniqueID; //8B
	private int pointer; //2B converted to short
	private int length; //2B converted to short	
	private Type type;

	private short[] route; //n*2
	private long[] time; //n*8

	public TraceRoutePacket(short[] trace, int size, Type type) {
		if (trace.length < 2) {
			throw new IllegalArgumentException(getClass().getName() + ": Trace length should be at least 2. [source, destination]");
		}
		this.length = trace.length;
		this.route = trace;
		this.time = new long[trace.length];
		this.pointer = 0;
		this.type = type;
		this.uniqueID = System.currentTimeMillis();
		if (getSize() < size) {
			data = new byte[size - getSize()];
			new Random().nextBytes(data);
		}

	}

	@Override
	public int getSize() {
		return 13 + route.length * 10 + (data==null? 0 :data.length);
	}

	@Override
	public int getHeaderSize() {
		return 0;
	}

	@Override
	public byte[] serialize() {
		ByteBuffer bb = ByteBuffer.allocate(getSize());

		bb.putLong(uniqueID);
		bb.put((byte) type.ordinal());
		bb.putShort((short) (pointer));
		bb.putShort((short) length);
		for (int i = 0; i < route.length; i++) {
			bb.putShort(route[i]);
			bb.putLong(time[i]);
		}
		bb.put(data);

		return bb.array();
	}

	@Override
	public boolean deserialize(ByteBuffer bb, int bLength) {
		bb.order(ByteOrder.BIG_ENDIAN);

		uniqueID = bb.getLong();
		type = Type.values()[bb.get() & 0xFF];
		pointer = bb.getShort() & 0xFFFF;
		length = bb.getShort() & 0xFFFF;
		if (bLength < 12 + length * 10) {
			return false;
		}
		route = new short[length];
		time = new long[length];
		for (int i = 0; i < length; i++) {
			route[i] = bb.getShort();
			time[i] = bb.getLong();
		}
		bLength -= 12 + 10 * length;
		data = new byte[bLength];
		bb.get(data);
		return true;
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = prime;
		result = prime * result + Arrays.hashCode(route);
		result = prime * result + (int) (uniqueID ^ (uniqueID >>> 32));
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (getClass() != obj.getClass())
			return false;
		TraceRoutePacket other = (TraceRoutePacket) obj;
		if (!Arrays.equals(route, other.route))
			return false;
		if (uniqueID != other.uniqueID)
			return false;
		return true;
	}

	@Override
	public String toString() {
		StringBuilder sb = new StringBuilder();
		sb.append(String.format("Trace: [%d] [%d > %d] {", uniqueID, getSrcAddr() & 0xFFFF, getDstAddr() & 0xFFFF));
		for (int i = 0; i < length; i++) {
			sb.append('[');
			sb.append(route[i]);
			sb.append(':');
			sb.append(time[i]);
			sb.append(']');
		}
		sb.append("}");
		return sb.toString();
	}

	@Override
	public short getDstAddr() {
		if (pointer < route.length) {
			return route[pointer];
		}
		return 0;
	}

	@Override
	public short getSrcAddr() {
		if (pointer > 0) {
			return route[pointer - 1];
		}
		return 0;
	}

	public boolean hasNext() {
		return pointer + 1 < route.length;
	}

	public boolean increment() {
		if (hasNext()) {
			time[pointer] = System.currentTimeMillis();
			pointer++;
			return true;
		}
		time[pointer] = System.currentTimeMillis();
		return false;
	}
	
	public boolean toRemember() {
		return type == Type.BUMERANG && pointer < route.length/2;
	}

	public TraceRoutePacket copy() {
		return new TraceRoutePacket(route, getSize(), type);
	}
	
	public String toStats() {
		StringBuilder sb = new StringBuilder();
		sb.append(uniqueID);
		sb.append('\t');
		for (int i = 0; i < length; i++) {
			sb.append(String.format("%d\t%d\t", route[i], time[i]));
		}
		try {
		if(type == Type.BUMERANG && pointer > (route.length)/2) {
			int ptr = pointer;
			while(ptr > route.length/2) {
				int optr = route.length-1-ptr;
				int dist = (ptr-optr)/2;
				int rtt = (int)(time[ptr]-time[optr]);
				double speed = ((float)(dist*2*getSize()))/((float)rtt/1000f);
				sb.append(String.format("[%d <-> %d = %d (%.2f Bps)]\t",route[ptr]&0xFFFF, route[optr]&0xFFFF, rtt, speed));
				ptr --;
			}
		}
		} catch (Exception e) {
			e.printStackTrace();
		}
		return sb.toString();
	}
	
	public long[] getTime() {
		return time;
	}
	

}
