package atc.safefly.communication.common.packet;

import java.io.Serializable;
import java.nio.ByteBuffer;
import java.nio.ByteOrder;

import atc.safefly.utils.SafeflyConstants.Cmd_Type;

public class VisualizationPacket extends Packet implements Serializable{

	private static final long serialVersionUID = 8308369417796841931L;
	/*************** Packet Factory registration ************/
	public static final short ID = 0x0002;
	static {
		PacketFactory.registerPacket(new VisualizationPacket());

	}

	@Override
	public Packet getInstance() {
		return new VisualizationPacket();
	}

	@Override
	public short getApplicationID() {
		return ID;
	}

	public VisualizationPacket() {
	}

	/*********************************************************/

	public Cmd_Type cmdType;
	public CmdID cmdId;

	public static final int HEADER_SIZE = 4;

	private VisualizationData visData;

	public VisualizationPacket(ByteBuffer bb, int length) {
		deserialize(bb, length);
	}

	@Override
	public boolean deserialize(ByteBuffer bb, int length) {
		if (length < HEADER_SIZE)
			return false;
		if (bb.get() == 0) {
			cmdType = Cmd_Type.CMDTYPE_REQ;
		} else {
			cmdType = Cmd_Type.CMDTYPE_REPLY;
		}
		try {
			cmdId = CmdID.values()[bb.get()];
		} catch (ArrayIndexOutOfBoundsException ex) {
			return false;
		}
		switch (cmdId) {
			case CMDID_SET_NON_COOP_POINT:
				visData = new CmdParSetNonCoopPoint();
				break;
			case CMDID_DEL_NON_COOP_POINT:
			case CMDID_DEL_COOP_PLAN:
				visData = new CmdParDelPoint();
				break;
			case CMDID_SET_COOP_PLAN:
				visData = new CmdParSetCoopPoint();
				break;
			case CMDID_SET_RANGE:
				throw new UnsupportedOperationException("Command ID 'SetRange' was not implemented yet in VisualizationPacket");
				//				break;
			case CMDID_STATUS:
				visData = new CmdParStatus();
				break;
			default:
				break;
		}
		visData.deserialize(bb, length - HEADER_SIZE);

		return true;
	}

	@Override
	public byte[] serialize() {
		if (visData == null)
			return null;

		ByteBuffer bb = ByteBuffer.allocate(HEADER_SIZE + visData.getSize());
		bb.order(ByteOrder.LITTLE_ENDIAN);
		bb.put(cmdType.getValue());
		bb.put((byte) cmdId.ordinal());
		bb.put((byte) 0);
		bb.put((byte) 0);
		bb.put(visData.serialize());
		return bb.array();
	}

	@Override
	public String toString() {
		StringBuilder builder = new StringBuilder();
		builder.append("VisualizationPacket [");
		if (cmdType != null)
			builder.append("cmdType=").append(cmdType).append(", ");
		if (cmdId != null)
			builder.append("cmdId=").append(cmdId).append(", ");
		if (visData != null)
			builder.append("visData=").append(visData);
		builder.append("]");
		return builder.toString();
	}

	/*-------------------------------------------------*/
	/*-------- Commands IDs  --------------------------*/
	/*-------------------------------------------------*/
	public enum CmdID {
		CMDID_NONE, CMDID_SET_NON_COOP_POINT, // nastaveni zobrazeni bodu v nekooperativnim modu
		CMDID_DEL_NON_COOP_POINT, // smazani bodu v nekooperativnim modu
		CMDID_SET_COOP_PLAN, // nastaveni zobrazeni planu v kooperativnim modu
		CMDID_DEL_COOP_PLAN, // smazani planu v kooperativnim modu
		CMDID_SET_RANGE, // nastaveni urceni rozsahu
		CMDID_STATUS; // zasilany status z Integry
	}

	public enum TrafficChange {
		LEVEL((byte) 0), DESCEND((byte) 1), ASCEND((byte) 2);

		private byte value;

		private TrafficChange(byte value) {
			this.value = value;
		}

		public byte getValue() {
			return value;
		}
	}

	public enum Color {
		GRAY((byte) 0b00000), RED((byte) 0b00100), GREEN((byte) 0b01000), YELLOW((byte) 0b01100), BLUE((byte) 0b10000), MAGENTA((byte) 0b10100), CYAN(
				(byte) 0b11000), WHITE((byte) 0b11100);

		private byte value;

		private Color(byte value) {
			this.value = value;
		}

		public byte getValue() {
			return value;
		}
	}

	public interface VisualizationData extends Serializable{
		public boolean deserialize(ByteBuffer bb, int length);

		public byte[] serialize();

		public int getSize();
	}

	public class CmdParSetNonCoopPoint implements VisualizationData {

		/**
		 * 
		 */
		private static final long serialVersionUID = 1989093990185833958L;
		private short planeId;
		private byte[] pointCoordinates;
		private byte flagsPoint;
		private byte[] tolerance;

		public static final int SIZE = 10;

		public CmdParSetNonCoopPoint() {
		}

		public CmdParSetNonCoopPoint(int planeId, int x, int y, int z, int ta, int td, int tp, int tz, Color color, TrafficChange change) {
			this.planeId = (short) planeId;
			pointCoordinates = new byte[3];
			pointCoordinates[0] = (byte) x;
			pointCoordinates[1] = (byte) y;
			pointCoordinates[2] = (byte) z;
			tolerance = new byte[4];
			tolerance[0] = (byte) ta;
			tolerance[1] = (byte) td;
			tolerance[2] = (byte) tp;
			tolerance[3] = (byte) tz;
			flagsPoint = 0;
			flagsPoint |= change.getValue();
			flagsPoint |= color.getValue();

		}

		@Override
		public boolean deserialize(ByteBuffer bb, int length) {
			if (length < SIZE)
				return false;
			bb.order(ByteOrder.LITTLE_ENDIAN);
			planeId = bb.getShort();
			pointCoordinates = new byte[3];
			bb.get(pointCoordinates);
			flagsPoint = bb.get();
			tolerance = new byte[4];
			bb.get(tolerance);
			return true;
		}

		@Override
		public byte[] serialize() {
			ByteBuffer bb = ByteBuffer.allocate(SIZE);
			bb.order(ByteOrder.LITTLE_ENDIAN);
			bb.putShort(planeId);
			bb.put(pointCoordinates);
			bb.put(flagsPoint);
			bb.put(tolerance);

			return bb.array();
		}

		@Override
		public int getSize() {
			return SIZE;
		}

	}

	public class CmdParDelPoint implements VisualizationData {

		/**
		 * 
		 */
		private static final long serialVersionUID = -8524442938783628275L;

		public static final int SIZE = 2;

		private short planeId;

		public CmdParDelPoint() {
		}

		public CmdParDelPoint(int planeId) {
			this.planeId = (short) planeId;
		}

		@Override
		public boolean deserialize(ByteBuffer bb, int length) {
			if (length < SIZE)
				return false;
			bb.order(ByteOrder.LITTLE_ENDIAN);
			planeId = bb.getShort();
			return true;
		}

		@Override
		public byte[] serialize() {
			ByteBuffer bb = ByteBuffer.allocate(SIZE);
			bb.order(ByteOrder.LITTLE_ENDIAN);
			bb.putShort(planeId);
			return bb.array();
		}

		@Override
		public int getSize() {
			return SIZE;
		}

	}

	public class CmdParSetCoopPoint implements VisualizationData {

		/**
		 * 
		 */
		private static final long serialVersionUID = 743951764703721115L;

		public static final int SIZE = 32;

		private short planeId;
		private byte pointsCnt;
		private byte flagsPlan;
		private byte[] points;

		public CmdParSetCoopPoint() {
		}

		public CmdParSetCoopPoint(int planeId, byte[] points, int planCount, boolean zeroPlane, boolean thisPlane, Color color) {
			if (points.length != 7 * 4)
				throw new IllegalArgumentException("Illegal count of points!");
			this.planeId = (short) planeId;
			this.pointsCnt = (byte) planCount;
			this.points = points;
			flagsPlan = 0;
			if (zeroPlane)
				flagsPlan |= 0x01;
			if (thisPlane)
				flagsPlan |= 0x02;
			flagsPlan |= color.getValue();
		}

		@Override
		public boolean deserialize(ByteBuffer bb, int length) {
			if (length < SIZE)
				return false;
			bb.order(ByteOrder.LITTLE_ENDIAN);
			planeId = bb.getShort();
			pointsCnt = bb.get();
			flagsPlan = bb.get();
			points = new byte[pointsCnt * 3];
			bb.get(points);
			// Skip zero positions
			bb.position(bb.position() + (3 * (7 - pointsCnt)));

			return true;
		}

		@Override
		public byte[] serialize() {
			ByteBuffer bb = ByteBuffer.allocate(SIZE);
			bb.order(ByteOrder.LITTLE_ENDIAN);
			bb.putShort(planeId);
			bb.put(pointsCnt);
			bb.put(flagsPlan);
			bb.put(points);
			return bb.array();
		}

		@Override
		public int getSize() {
			return SIZE;
		}

		@Override
		public String toString() {
			StringBuilder builder = new StringBuilder();
			builder.append("CmdParSetCoopPoint [planeId=").append(planeId).append(", pointsCnt=").append(pointsCnt).append(", flagsPlan=").append(flagsPlan)
					.append(", ");
			if (points != null) {
				builder.append("points=");
				for (int i = 0; i < points.length; i++) {
					if (i == 0) {
						builder.append("[");
					} else if (i == points.length - 1) {
						builder.append("]");
					} else if (i % 4 == 0) {
						builder.append("], [");
					} else {
						builder.append(", ");
					}
					builder.append(points[i] & 0xFF);
				}
			}
			builder.append("]");
			return builder.toString();
		}

	}

	public class CmdParStatus implements VisualizationData {

		/**
		 * 
		 */
		private static final long serialVersionUID = 2319125048202062907L;

		public static final int SIZE = 20;

		public long applicationProtocolError;
		public long transportProtocolError;
		public long setPointCommandCount;
		public long setPlanCommandCount;
		public int activePointsCount;
		public int activePlansCount;
		public int currentScale;

		public CmdParStatus() {
		}

		@Override
		public boolean deserialize(ByteBuffer bb, int length) {
			if (length < SIZE)
				return false;
			bb.order(ByteOrder.LITTLE_ENDIAN);
			applicationProtocolError = bb.getInt() & 0xFFFFFFFF;
			transportProtocolError = bb.getInt() & 0xFFFFFFFF;
			setPointCommandCount = bb.getInt() & 0xFFFFFFFF;
			setPlanCommandCount = bb.getInt() & 0xFFFFFFFF;
			activePointsCount = bb.get() & 0xFF;
			activePlansCount = bb.get() & 0xFF;
			currentScale = bb.get() & 0xFF;

			return true;
		}

		@Override
		public byte[] serialize() {
			ByteBuffer bb = ByteBuffer.allocate(SIZE);
			bb.order(ByteOrder.LITTLE_ENDIAN);
			bb.putInt((int) applicationProtocolError);
			bb.putInt((int) transportProtocolError);
			bb.putInt((int) setPointCommandCount);
			bb.putInt((int) setPlanCommandCount);
			bb.put((byte) activePointsCount);
			bb.put((byte) activePlansCount);
			bb.put((byte) currentScale);
			return bb.array();
		}

		@Override
		public int getSize() {
			return SIZE;
		}

		@Override
		public String toString() {
			return String
					.format("CmdParStatus: \napplicationProtocolError=\t%s\ntransportProtocolError=\t%s\nsetPointCommandCount=\t%s\nsetPlanCommandCount=\t%s\nactivePointsCount=\t%s\nactivePlansCount=\t%s\ncurrentScale=\t%s",
							applicationProtocolError, transportProtocolError, setPointCommandCount, setPlanCommandCount, activePointsCount, activePlansCount,
							currentScale);
		}

	}

	public static VisualizationPacket getDeletePacket(int planeId, boolean cooperative) {
		VisualizationPacket packet = new VisualizationPacket();
		packet.cmdType = Cmd_Type.CMDTYPE_REQ;
		if (cooperative) {
			packet.cmdId = CmdID.CMDID_DEL_COOP_PLAN;
		} else {
			packet.cmdId = CmdID.CMDID_DEL_NON_COOP_POINT;
		}
		packet.visData = packet.new CmdParDelPoint(planeId);
		return packet;
	}

	public static VisualizationPacket getCoopPlanPacket(int planeId, byte[] points, int planCount, boolean withPlane, boolean thisPlane, Color color) {
		VisualizationPacket packet = new VisualizationPacket();
		packet.cmdType = Cmd_Type.CMDTYPE_REQ;
		packet.cmdId = CmdID.CMDID_SET_COOP_PLAN;
		packet.visData = packet.new CmdParSetCoopPoint(planeId, points, planCount, withPlane, thisPlane, color);
		return packet;
	}

	public static VisualizationPacket getNonCoopPacket(int planeId, int x, int y, int z, int ta, int td, int tp, int tz, Color color, TrafficChange change) {
		VisualizationPacket packet = new VisualizationPacket();
		packet.cmdType = Cmd_Type.CMDTYPE_REQ;
		packet.cmdId = CmdID.CMDID_SET_NON_COOP_POINT;
		packet.visData = packet.new CmdParSetNonCoopPoint(planeId, x, y, z, ta, td, tp, tz, color, change);
		return packet;
	}

}
