package atc.safefly.communication.common.serial;

import java.io.File;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.net.ServerSocket;

import atc.safefly.communication.utils.StreamPipeThread;

public class UniversalSerialConnector extends AbstractSerialConnector {
	public static final int B115200 = 115200;
	public static final int B230400 = 230400;
	public static final int B9600 = 9600;
	public static final int B57600 = 57600;

	private Process p;
	private File ex;

	Thread errorPipe;
	Thread consolePipe;

	private String serialName;
	private int baudRate;
	private int debugLevel;

	private int port;

	private SocketSerialConnector subConnector;

	public UniversalSerialConnector(String serialName, int baudRate, int debugLevel) {
		super();
		this.serialName = serialName;
		this.baudRate = baudRate;
		this.debugLevel = debugLevel;
		isConnected = false;
	}

	public UniversalSerialConnector(String serialName, int baudRate) {
		this(serialName, baudRate, 0);
	}

	@Override
	public boolean connect() {
		if (!isConnected()) {
			// ---------------Get free port---------------
			try {
				ServerSocket ss = new ServerSocket(0);
				port = ss.getLocalPort();
				ss.close();
				ss = null;
			} catch (IOException e) {
				e.printStackTrace();
				return false;
			}

			subConnector = new SocketSerialConnector(port);
			// -------------------------------------------

			String command;
			if (System.getProperty("os.name").startsWith("Windows")) {
				if (!serialName.startsWith("COM")) {
					System.out.println(serialName + " is not valid Windows serial port name!");
					return false;
				}
				if (extractFile("serialtoip.exe")) {
					// Windows App
					command = String.format("serialtoip.exe --serialport %s --baudrate %d --socketport %d", serialName, baudRate, port);
				} else {
					return false;
				}
			} else {
				// Unix system / probably gumstix
				if (serialName.startsWith("tty")) {
					serialName = "/dev/" + serialName;
				}
				if (!serialName.startsWith("/dev/tty")) {
					System.out.println(serialName + " is not valid Unix serial port name!");
					return false;
				}
				if (extractFile("sc")) {
					// Unix app
					command = String.format("./sc -p %d -s %s -b %d -g %d", port, serialName, baudRate, debugLevel);
				} else {
					return false;
				}
			}

			// Run underlying process and pipe input streams
			try {
				System.out.println(command);
				p = Runtime.getRuntime().exec(command);
				// -------- Consume the output stream of the process
				OutputStream output = null;
				if (debugLevel > 0) {
					output = System.out;
				}
				consolePipe = new StreamPipeThread(p.getInputStream(), output);
				consolePipe.start();

				errorPipe = new StreamPipeThread(p.getErrorStream(), output);
				errorPipe.start();
				// ----------------------------------------------------------------
			} catch (IOException e) {
				e.printStackTrace();
				return false;
			}
			try {
				Thread.sleep(500);
			} catch (InterruptedException e) {
				return false;
			}
			if (subConnector.connect()) {
				in = subConnector.getInputStream();
				out = subConnector.getOutputStream();
				isConnected = true;
			} else {
				// Connection failed, dispose everything
				dispose();
			}
		}
		return isConnected;
	}

	private boolean extractFile(String fileName) {
		ex = new File(fileName);
		if (!ex.exists()) {
			try {
				InputStream src = ClassLoader.getSystemResourceAsStream(fileName);
				if (src == null) {
					System.err.println("Cannot load resource \"" + fileName + "\"");
					return false;
				}
				FileOutputStream out = new FileOutputStream(ex);
				byte[] buffer = new byte[1024];
				int rc;
				while ((rc = src.read(buffer)) > 0) {
					out.write(buffer, 0, rc);
				}
				src.close();
				out.close();
				ex.setExecutable(true);
			} catch (Exception e) {
				e.printStackTrace();
				return false;
			}
		}
		return ex.exists();
	}

	@Override
	public void dispose() {
		if (subConnector != null) {
			subConnector.dispose();
			subConnector = null;
		}
		if (consolePipe != null) {
			consolePipe.interrupt();
		}
		if (errorPipe != null) {
			errorPipe.interrupt();
		}
		if (p != null) {
			p.destroy();
			try {
				p.waitFor();
				p = null;
			} catch (InterruptedException e) {
				// DO nothing
			}
		}
		if (ex != null) {
			ex.delete();
		}
		isConnected = false;
		restoreStreams();
	}

	@Override
	protected void finalize() throws Throwable {
		dispose();
		super.finalize();
	}

	@Override
	public String getPortName() {
		return serialName+"@"+baudRate;
	}
	
}
