package atc.safefly.communication.modem;

import java.io.IOException;
import java.nio.ByteBuffer;

import atc.procerus.utils.ByteUtils;
import atc.safefly.communication.common.packet.PacketConnectionOwner;
import atc.safefly.communication.common.packet.PacketConnector;
import atc.safefly.communication.common.packet.PacketFactory;
import atc.safefly.communication.common.packet.PacketInterface;
import atc.safefly.communication.common.serial.SerialConnector;
import atc.safefly.utils.SafeflyConstants;

public class DataPacketConnector extends PacketConnector {

	/**
	 * ****************** Packet structure ***
	 * *************************************** ** StartByte | PacketData |
	 * Checksum ** ***************************************
	 */

	// /////////////////////////////////
	// //////// Escape characters///////
	// /////////////////////////////////
	public static final byte START = (byte) 0x7E;
	public static final byte ESC = (byte) 0x7D;
	public static final byte XON = (byte) 0x11;
	public static final byte XOFF = (byte) 0x13;

	private static final int BUFFER_SIZE = 256;
	private static final int MAX_PACKET_ZIZE = 255;

	public DataPacketConnector(SerialConnector subConnector, PacketConnectionOwner owner) {
		super(subConnector, owner);
	}

	public DataPacketConnector(SerialConnector subConnector) {
		this(subConnector, null);
	}

	protected void readPacket() {
		// Receive buffer
		byte[] buffer = new byte[BUFFER_SIZE];
		int ptr = 0;
		// Decoded packet buffer
		ByteBuffer bb = ByteBuffer.allocateDirect(BUFFER_SIZE);
		// State indicators
		boolean startReceived = false;
		boolean lastEscape = false;
		int length = -1;
		short type = 0;
		int crcChecksum = 0;
		int size = 0;
		/******************** Read data from stream ***************/
		while (!Thread.interrupted()) {
			// Read data to buffer
			try {
				if (lastEscape) {
					// Last buffered character was ESC, so add ESC at the
					// beginning of next buffer
					buffer[0] = ESC;
					size = in.read(buffer, 1, buffer.length - 1);
					size++;
					lastEscape = false;
				} else {
					size = in.read(buffer);
				}
				ptr = 0;
			} catch (IOException e) {
				System.err.println("Reading of bytes failed: " + getClass().getName());
				return;
			}

			/****************** Receive packet cycle **************/
			while (ptr < size) {
				/************** Find first START byte *************/
				while (!startReceived && ptr < size) {
					switch (buffer[ptr]) {
					case START:
						startReceived = true;
						bb.clear();
						crcChecksum = 0;
						length = -1;
						break;
					case ESC:
						// Don't escape, but just skip
						ptr++;
						if (ptr == size) {
							lastEscape = true;
						}
						break;
					case XON:
						System.out.println("XON received");
						break;
					case XOFF:
						System.out.println("XOFF received");
						break;
					default:
						System.out.println(getClass().getName() + "> Data skipped");
					}
					ptr++;
				}
				/**************************************************/

				/************ Read packet header ******************/
				while (bb.position() < 4 && ptr < size) {
					switch (buffer[ptr]) {
					case START:
						System.err.println(getClass().getName() + "> Packet broken by START byte while reading length!");
						break;
					case XON:
						System.out.println("XON received");
						break;
					case XOFF:
						System.out.println("XOFF received");
						break;
					case ESC:
						ptr++;
						if (ptr == size) {
							lastEscape = true;
							break;
						} else {
							// Escape next byte
							buffer[ptr] ^= 0x20;
						}
						// No break to process escaped byte with next
						// instruction
					default:
						bb.put(buffer[ptr]);
						crcChecksum += buffer[ptr];
					}
					ptr++;
					if(bb.position() == 4) {
						bb.position(0);
						//Header received, decode length and type
						length = bb.getShort() & 0xFFFF;
						type = bb.getShort();
						// 2 Extra Bytes for CRC
						length += 2;
					}
				}
				/*************************************************/
				
				/*********** Read packet data ********************/
				while (length > 0 && ptr < size) {
					switch (buffer[ptr]) {
					case START:
						System.err.println(getClass().getName() + "> Packet broken by START byte while reading data!");
						bb.clear();
						crcChecksum = 0;
						length = -1;
						break;
					case XON:
						System.out.println("XON received");
						break;
					case XOFF:
						System.out.println("XOFF received");
						break;
					case ESC:
						ptr++;
						if (ptr == size) {
							lastEscape = true;
							break;
						} else {
							// Escape next byte
							buffer[ptr] ^= 0x20;
						}
						// No break to process escaped byte with next
						// instruction
					default:
						bb.put(buffer[ptr]);
						if (length > 2) {
							// CRC computation
							crcChecksum = (crcChecksum >>> 8) ^ SafeflyConstants.CRC_LOOKUP_TABLE[(crcChecksum ^ buffer[ptr]) & 0xff];
						}
						length--;
					}
					ptr++;
				}

				// Complete packet received
				if (length == 0) {
					// Save and reset buffer position
					int bufferSize = bb.position() - 2;
					bb.position(bufferSize);
					int packetCrc = bb.getShort() & 0xFFFF;
					bb.position(0);

					if (crcChecksum == packetCrc) {
						try {
							PacketInterface packet = PacketFactory.getPacket(type);
							packet.deserialize(bb, length);
							packetQueue.put(packet);
						} catch (InterruptedException e) {
							return;
						} catch (NullPointerException e) {
							/** Do nothing */
						}
					} else {
						System.out.println(getClass().getName() + "> Packet checksum failed.");

					}
					// Wait for next START byte
					startReceived = false;
				}
			}

			if (size < 0) {
				// If size is -1 EndOfStream is indicated by read() method
				System.err.println(getClass().getName() + "> Nothing to read from stream");
				return;
			}
		}
	}

	/**
	 * Escape data to send, compute checksum and add header
	 * 
	 * @return true if packet sent correctly, false otherwise
	 */
	@Override
	public synchronized boolean sendPacket(PacketInterface packet) {
		if (isConnected()) {
			byte[] sPacket = packet.serialize();
			int crcChecksum = 0;
			// 2B type + 2B length
			ByteBuffer header = ByteBuffer.allocate(4);
			header.putShort(packet.getApplicationID());
			header.putShort((short) sPacket.length);

			// Allocate space for escaping data and header
			ByteBuffer buffer = ByteBuffer.allocate((2 * (sPacket.length + 3)));
			buffer.put(START);
			/*************** Write escaped packet header ************/
			// Packet length + 2B for crcChecksum

			header.position(0);
			while (header.hasRemaining()) {
				byte bVal = header.get();
				crcChecksum = (crcChecksum >>> 8) ^ SafeflyConstants.CRC_LOOKUP_TABLE[(crcChecksum ^ bVal) & 0xff];
				switch (bVal) {
				case (START):
				case (ESC):
				case (XON):
				case (XOFF):
					buffer.put(ESC);
					buffer.put((byte) (bVal ^ 0x20));
					break;
				default:
					buffer.put(bVal);
				}
			}
			/*************** Write escaped packet data ************/
			for (int i = 0; i < sPacket.length; i++) {
				// Compute CRC checksum
				crcChecksum = (crcChecksum >>> 8) ^ SafeflyConstants.CRC_LOOKUP_TABLE[(crcChecksum ^ sPacket[i]) & 0xff];
				// Escape and write
				switch (sPacket[i]) {
				case (START):
				case (ESC):
				case (XON):
				case (XOFF):
					buffer.put(ESC);
					buffer.put((byte) (sPacket[i] ^ 0x20));
					break;
				default:
					buffer.put(sPacket[i]);
				}
			}
			/*************** Write escaped checksum ************/
			byte[] crc = ByteUtils.shortToBytes((short) crcChecksum);
			for (int i = 0; i < crc.length; i++) {
				switch (crc[i]) {
				case (START):
				case (ESC):
				case (XON):
				case (XOFF):
					buffer.put(ESC);
					buffer.put((byte) (crc[i] ^ 0x20));
					break;
				default:
					buffer.put(crc[i]);
				}
			}

			/************* Write data into stream *************/
			try {
				byte[] pktBytes = new byte[buffer.position()];
				buffer.position(0);
				buffer.get(pktBytes);

				// System.out.println(Arrays.toString(pktBytes));

				out.write(pktBytes);
				out.flush();
			} catch (IOException e) {
				e.printStackTrace();
				return false;
			}
			return true;
		}
		return false;
	}
	
	@Override
	public int getMaxPacketSize() {
		return MAX_PACKET_ZIZE;
	}
}
