package atc.safefly.communication.modem;

import java.io.IOException;
import java.net.SocketException;
import java.net.SocketTimeoutException;
import java.nio.ByteBuffer;
import java.util.concurrent.TimeUnit;

import atc.safefly.communication.common.packet.Packet;
import atc.safefly.communication.common.packet.PacketConnector;
import atc.safefly.communication.common.packet.PacketInterface;
import atc.safefly.communication.common.serial.SerialConnector;

public class MicrohardDiagnosticPacketConnector extends PacketConnector {

	private static final byte[] NET_CONF = new byte[] { 3, 0, 1, 45 };
	private static final byte ID_RSSI = (byte) 105;

	private static int timeOut = 500; // ms

	public MicrohardDiagnosticPacketConnector(SerialConnector subConnector) {
		super(subConnector, null);
	}

	public synchronized byte[] getDiagnostic(int addr) {
		ByteBuffer bb = ByteBuffer.allocate(4);
		bb.putInt(addr);
		packetQueue.clear();
		sendPacket(new Packet(new byte[] { 3, bb.get(2), bb.get(3), 1 }));

		PacketInterface response;
		try {
			response = packetQueue.poll(timeOut, TimeUnit.MILLISECONDS);
		} catch (InterruptedException e) {
			return null;
		}
		return response.getData();
	}

	public synchronized byte[] getParametrsByID(int[] ID, int addr) {
		ByteBuffer bb = ByteBuffer.allocate(4);
		bb.putInt(addr);
		byte[] packet = new byte[4 + ID.length];
		packet[0] = (byte) (2 + ID.length);
		packet[1] = bb.get(2);
		packet[2] = bb.get(3);
		packet[3] = 20;
		for (int i = 0; i < ID.length; i++) {
			packet[i + 4] = (byte) ID[i];
		}
		packetQueue.clear();
		sendPacket(new Packet(packet));
		PacketInterface response;
		try {
			response = packetQueue.poll(timeOut, TimeUnit.MILLISECONDS);
		} catch (InterruptedException e) {
			return null;
		}
		if(response == null) {
			return null;
		}
		return response.getData();
	}

	/**
	 * Response format : [0] - packet size [1-2] - Master UA [3] - Response ID
	 * {[2B]}* - Other unit addresses in network
	 **/

	public synchronized byte[] getNetworkConfiguration() {
		PacketInterface response;
		// send broadcast RSSI packet to update network map in master
		getParametrsByID(new int[] { ID_RSSI }, 65535);
		try {
			// wait for all responses
			Thread.sleep(2 * timeOut);
			// clear all responses
			packetQueue.clear();
			// get network configuration
			sendPacket(new Packet(NET_CONF));
			response = packetQueue.poll(timeOut, TimeUnit.MILLISECONDS);
		} catch (InterruptedException e) {
			return null;
		}
		return response.getData();
	}

	@Override
	public synchronized boolean sendPacket(PacketInterface packet) {
		if (!packet.getClass().equals(atc.safefly.communication.common.packet.Packet.class)) {
			System.err.println(getClass().getName() + "> Unsupported packet type: " + packet.getClass().getName());
			return false;
		}
		if (isConnected()) {
			try {
				out.write(packet.serialize());
				out.flush();
				return true;
			} catch (IOException e) {
				e.printStackTrace();
			}
		}
		return false;
	}

	protected void readPacket() {
		while (!Thread.interrupted()) {
			byte[] buff = null;
			try {
				int length;
				if ((length = in.read()) > -1) {
					buff = new byte[length];
					int val = 0;
					for (int i = 0; i < buff.length; i++) {
						if ((val = in.read()) == -1)
							return;
						buff[i] = (byte) val;
					}
				}
				Packet p = new Packet(buff);
				packetQueue.put(p);

			} catch (InterruptedException e) {
				e.printStackTrace();
				return;
			} catch (SocketTimeoutException e) {
				continue;
			} catch (SocketException e) {
				continue;
			} catch (IOException e) {
				e.printStackTrace();
				return;
			}
		}

	}

	@Override
	public boolean open() {
		super.open();
		// Use own push method as methods response
		pushingThread.interrupt();
		return true;
	}
	
	@Override
	public int getMaxPacketSize() {
		return Integer.MAX_VALUE;
	}

}
