package atc.safefly.communication.modem;

import java.io.IOException;
import java.net.InetAddress;
import java.net.SocketTimeoutException;
import java.net.UnknownHostException;
import java.util.Scanner;
import java.util.concurrent.TimeUnit;

import atc.safefly.communication.common.packet.Packet;
import atc.safefly.communication.common.packet.PacketConnector;
import atc.safefly.communication.common.packet.PacketInterface;
import atc.safefly.communication.common.serial.SerialConnector;
import atc.safefly.communication.common.serial.SocketSerialConnector;
import atc.safefly.communication.common.serial.UniversalSerialConnector;

public class TerminalConnector extends PacketConnector {

	private static final byte[] NL = "\r\n".getBytes();

	private static final String HELP = "Usage: [COM/tty_port Baudrate] | [tcp_port IP_address]";

	private static final long lineTiomeout = 500;

	private static String[] commonConfig = new String[] { "ATS102=0", // Serial
																		// Baud
																		// Rate
																		// 0-230400/1-115200/7-9600
			"ATS103=3", // Wireless Link Rate 2-172800/3-230400/(8-1.2 Mbps)
			"ATS104=147258369", // Network address
			"ATS107=AtgSafeFlySecure", // Static data mask (up to 16 chars)
			"ATS108=30", // Set maximum output power 20-30(1W)
			"ATS111=10", // Set minimal packet size
			"ATS112=255", // Packet Max Size (1-256)
			"ATS113=0", // Packet retransmission set to 0
			"ATS116=10", // Set maximal inter-character timeout
	};

	public static void main(String[] args) {
		InetAddress adr;
		try {
			adr = InetAddress.getByAddress(new byte[] { 127, 0, 0, 1 });
		} catch (UnknownHostException e1) {
			e1.printStackTrace();
			return;
		}
		SerialConnector conn;
		try {
			if (args.length > 0) {
				if (args.length == 2) {
					if (args[0].startsWith("COM") || args[0].startsWith("/dev/tty")) {
						conn = new UniversalSerialConnector(args[0], Integer.parseInt(args[1]));
					} else {
						String[] sAdr = args[1].split(",");
						byte[] bAdr = new byte[sAdr.length];
						for (int i = 0; i < sAdr.length; i++) {
							bAdr[i] = (byte) Integer.parseInt(sAdr[i]);
						}
						adr = InetAddress.getByAddress(bAdr);
						conn = new SocketSerialConnector(adr, Integer.parseInt(args[0]));
					}
				} else {
					conn = new SocketSerialConnector(adr, Integer.parseInt(args[0]));
				}
			} else {
				System.out.println(HELP);
				return;
			}
		} catch (NumberFormatException e) {
			System.out.println(HELP);
			return;
		} catch (UnknownHostException e) {
			e.printStackTrace();
			System.out.println(HELP);
			return;
		}

		TerminalConnector tc = new TerminalConnector(conn);
		if (tc.connect()) {
			System.out.print(tc.readLines());
		} else {
			System.err.println("TerminalConnector: Connection failed!");
			return;
		}

		Scanner sc = new Scanner(System.in);
		String line = sc.nextLine();
		while (!(line.equalsIgnoreCase("ATO") || line.equalsIgnoreCase("ATA") || line.equalsIgnoreCase("QUIT"))) {
			String answer = null;
			if (line.equalsIgnoreCase("setPtmMaster")) {
				tc.setP2mMaster();
				answer = tc.readLines();
			} else if (line.equalsIgnoreCase("setPtmSlave")) {
				tc.setP2mSlave(readPositiveInteger(sc));
				answer = tc.readLines();
			} else if (line.equalsIgnoreCase("setEteSlave")) {
				tc.setE2eSlave(readPositiveInteger(sc));
				answer = tc.readLines();
			} else if (line.equalsIgnoreCase("setEteMaster")) {
				tc.setE2eMaster();
				answer = tc.readLines();
			} else if (line.equalsIgnoreCase("view")) {
				tc.sendPacket("AT&V");
				answer = tc.readLines();
			} else if (line.equalsIgnoreCase("write")) {
				tc.sendPacket("ATW");
				answer = tc.readLines();
			} else if (line.equalsIgnoreCase("getaddress")) {
				answer = Integer.toString(tc.getAddress());
			} else if (line.equalsIgnoreCase("setaddress")) {
				tc.setAddress(readPositiveInteger(sc));
				answer = tc.readLines();
			} else {
				tc.sendPacket(line);
				answer = tc.readLines();
			}
			if (answer != null) {
				System.out.print(answer);
			} else {
				System.out.println("NO ANSWER");
			}
			line = sc.nextLine();
		}
		sc.close();
		tc.dispose();
	}

	private static int readPositiveInteger(Scanner sc) {
		while (true) {
			System.out.print("Slave adress=");
			try {
				int number = Integer.parseInt(sc.nextLine());
				if (number < 2 || number > 65534) {
					System.out.println("Error. Type integer address 2-65534");
					continue;
				}
				return number;
			} catch (NumberFormatException e) {
				System.out.println("Error. Type integer address.");
				continue;
			}
		}
	}

	public TerminalConnector(SerialConnector subConnector) {
		super(subConnector, null);
	}

	public void close() {
		if (isConnected()) {
			sendPacket("ATO");
			String answer = readLines();
			if (answer != null) {
				System.out.println(answer);
			} else {
				System.out.println("EXIT: NO ANSWER");
			}

			try {
				Thread.sleep(1000);
			} catch (InterruptedException e) {
			}
		}
		readThread.interrupt();
		packetQueue.clear();
		packetQueue = null;
		in = null;
		out = null;
		subConnector.restoreStreams();
		isConnected = false;
	}

	public int getAddress() {
		sendPacket("ATS105?");
		String answer = readLines();
		int address = -1;
		if (answer != null) {
			try {

				address = Integer.parseInt(answer.split("\r\n")[1]);
			} catch (NumberFormatException e) {
				System.err.println("Modem address returned incorrectly!");
			}
		}
		return address;
	}

	@Override
	protected boolean open() {
		super.open();
		// Using own push method readLines();
		pushingThread.interrupt();

		// Switch to AT command mode
		try {
			out.write("+++".getBytes());
			out.flush();
			Thread.sleep(1100);
		} catch (IOException e) {
			e.printStackTrace();
			dispose();
			return false;
		} catch (InterruptedException e) {
			e.printStackTrace();
			dispose();
			return false;
		}
		System.out.print(readLines());
		return true;

	}

	private String readLines() {
		StringBuffer sb = new StringBuffer();
		while (true) {
			String line;
			try {
				PacketInterface pi = packetQueue.poll(lineTiomeout, TimeUnit.MILLISECONDS);
				if (pi == null) {
					return sb.toString();
				}
				line = new String(pi.getData());
			} catch (InterruptedException e) {
				return sb.toString();
			}
			sb.append(line);
		}
	}

	@Override
	public boolean sendPacket(PacketInterface packet) {
		if (!packet.getClass().equals(Packet.class)) {
			System.err.println(getClass().getName() + "> Unsupported packet type:" + packet.getClass());
			return false;
		}
		if (isConnected()) {
			try {
				out.write(packet.serialize());
				out.write(NL);
				out.flush();
			} catch (IOException e) {
				e.printStackTrace();
				return false;
			}
			return true;
		}
		return false;
	}

	private synchronized boolean sendPacket(String line) {
		return sendPacket(new Packet(line.getBytes()));
	}

	public String setAddress(int address) {
		sendPacket("ATS105=" + address);
		return readLines();
	}

	public String setE2eMaster() {
		if (!isConnected()) {
			System.out.println("Not connected");
			return null;
		}
		try {
			// load PMP master defaults
			sendPacket("AT&F1");
			Thread.sleep(500);
			System.out.println(readLines());
			// Set network type to Everyone-to-Everyone
			sendPacket("ATS133=2");
			Thread.sleep(500);
			System.out.println(readLines());

			// TODO: Consider remove this next configuration
			// // Set Master Channel Request Timeout to one Time Slot
			// sendPacket("ATS234=1");
			// Thread.sleep(500);
			// // Set Routing Time to Live to 1000 slots (65535/Infinity)
			// sendPacket("ATS235=10");
			// Thread.sleep(500);
			System.out.println("CommonConf");
			for (int i = 0; i < commonConfig.length; i++) {
				sendPacket(commonConfig[i]);
				Thread.sleep(500);
				System.out.println(readLines());
			}

			sendPacket("AT&W");
			Thread.sleep(500);
			System.out.println(readLines());
			sendPacket("AT&V");
			System.out.println(readLines());

		} catch (InterruptedException e) {
			return null;
		}
		return readLines();
	}

	public String setE2eSlave(int address) {
		if (!isConnected())
			return null;

		try {
			// load PMP slave defaults
			sendPacket("AT&F2");
			Thread.sleep(500);
			System.out.println(readLines());
			// Set network type to Everyone-to-Everyone
			sendPacket("ATS133=2");
			Thread.sleep(500);
			System.out.println(readLines());
			// Set slave address
			sendPacket("ATS105=" + address);
			Thread.sleep(500);
			System.out.println(readLines());
			// Set destination address to broadcat
			sendPacket("ATS140=65535");
			Thread.sleep(500);
			System.out.println(readLines());

			for (int i = 0; i < commonConfig.length; i++) {
				sendPacket(commonConfig[i]);
				Thread.sleep(500);
				System.out.println(readLines());
			}

			sendPacket("AT&W");
			Thread.sleep(500);
			System.out.println(readLines());
			sendPacket("AT&V");
			System.out.println(readLines());

		} catch (InterruptedException e) {
			return null;
		}
		return readLines();
	}

	public String setP2mMaster() {
		if (!isConnected())
			return null;

		try {
			// load PMP master defaults
			sendPacket("AT&F1");
			Thread.sleep(500);
			System.out.println(readLines());
			// Set Master Channel Request Timeout to one Time Slot
			sendPacket("ATS234=1");
			Thread.sleep(500);
			System.out.println(readLines());
			// Set Routing Time to Live to 1000 slots (65535/Infinity)
			sendPacket("ATS235=100");
			Thread.sleep(500);
			System.out.println(readLines());

			sendPacket("ATS244=4"); // Channel Request Mode 4-Adaptive TDMA
			Thread.sleep(500);
			System.out.println(readLines());
			sendPacket("ATS115=1"); // Repeat Interval (Hop interval 1 for TDMA)
			Thread.sleep(500);
			System.out.println(readLines());
			for (int i = 0; i < commonConfig.length; i++) {
				sendPacket(commonConfig[i]);
				Thread.sleep(500);
				System.out.println(readLines());
			}

			sendPacket("AT&W");
			Thread.sleep(500);
			System.out.println(readLines());
			sendPacket("AT&V");
			System.out.println(readLines());

		} catch (InterruptedException e) {
			return null;
		}
		return readLines();
	}

	public String setP2mSlave(int address) {
		if (!isConnected())
			return null;

		try {
			// load PMP slave defaults
			sendPacket("AT&F2");
			Thread.sleep(500);
			System.out.println(readLines());
			// Set slave address
			sendPacket("ATS105=" + address);
			Thread.sleep(500);
			System.out.println(readLines());

			// Set destination address to broadcast
			sendPacket("ATS140=65535");
			Thread.sleep(500);
			System.out.println(readLines());

			sendPacket("ATS244=4"); // Channel Request Mode 4-Adaptive TDMA
			Thread.sleep(500);
			System.out.println(readLines());
			sendPacket("ATS115=1"); // Repeat Interval (Hop interval 1 for TDMA)
			Thread.sleep(500);
			System.out.println(readLines());
			for (int i = 0; i < commonConfig.length; i++) {
				sendPacket(commonConfig[i]);
				Thread.sleep(500);
				System.out.println(readLines());
			}

			sendPacket("AT&W");
			Thread.sleep(500);
			System.out.println(readLines());
			sendPacket("AT&V");
			System.out.println(readLines());

		} catch (InterruptedException e) {
			return null;
		}
		return readLines();
	}

	protected void readPacket() {
		while (!Thread.interrupted()) {
			try {
				int val = in.read();
				StringBuilder sb = new StringBuilder();
				while (val > 0) {
					char cVal = (char) val;
					if (cVal == '\n') {
						sb.append(cVal);
						break;
					}
					sb.append(cVal);
					val = in.read();
				}
				Packet p = new Packet(sb.toString().getBytes());
				packetQueue.put(p);
			} catch (InterruptedException e) {
				return;
			} catch (SocketTimeoutException e) {
				continue;
			} catch (IOException e) {
				return;
			} 
		}
	}

	@Override
	public int getMaxPacketSize() {
		return Integer.MAX_VALUE;
	}

}
