package atc.safefly.communication.utils;

import java.util.LinkedList;
import java.util.concurrent.LinkedBlockingQueue;
import java.util.concurrent.ThreadPoolExecutor;
import java.util.concurrent.TimeUnit;

import org.apache.commons.lang.NotImplementedException;

import atc.safefly.communication.common.packet.PacketInterface;

public class MessageRecord implements Comparable<MessageRecord> {

	public static enum State {
		NEW, SENDING, SENT, DST_UNREACHABLE, EXPIRED, FAILED
	}

	private PacketInterface packet;

	private LinkedList<MessageRecordHandler> handlers;

	// TODO hide fields
	private short dstAdr;
	private short srcAdr;

	private long enqueueTime;
	private long timeout;
	private long timeoutTime;

	private State state;

	/**
	 * @param message
	 * @param dstAdr
	 * @param srcAdr
	 * @param timeout
	 */
	public MessageRecord(PacketInterface packet, short dstAdr, short srcAdr, long timeout) {
		this.handlers = new LinkedList<MessageRecordHandler>();
		this.packet = packet;

		this.dstAdr = dstAdr;

		this.srcAdr = srcAdr;

		this.timeout = timeout;
		this.enqueueTime = System.currentTimeMillis();
		this.timeoutTime = enqueueTime + this.timeout;
		this.state = State.NEW;
	}

	/**
	 * @return the enqueueTime
	 */
	public long getEnqueueTime() {
		return enqueueTime;
	}

	/**
	 * @return the timeout
	 */
	public long getTimeout() {
		return timeout;
	}

	public boolean isExpired() {
		if ((timeoutTime - System.currentTimeMillis()) < 0) {
			return true;
		}
		return false;
	}

	/**
	 * @return packed wrapped into this MessageRecord
	 */
	public PacketInterface getPacket() {
		return packet;
	}

	/**
	 * @return the dstAdr
	 */
	public short getDstAdr() {
		return dstAdr;
	}

	/**
	 * @return the srcAdr
	 */
	public short getSrcAdr() {
		return srcAdr;
	}

	@Override
	public int compareTo(MessageRecord o) {
		if (enqueueTime < o.enqueueTime) {
			return -1;
		} else if (enqueueTime > o.enqueueTime) {
			return 1;
		}
		return 0;
	}

	public synchronized void setState(final State state) {

		if (!this.state.equals(state)) {
			System.out.println(enqueueTime+">>>"+this.state.name()+"->"+state.name());
			this.state = state;

			final MessageRecord toHandle = this;
			// Start anonymous thread to call all handlers
			new Thread(new Runnable() {

				@Override
				public void run() {
					for (final MessageRecordHandler handler : handlers) {
						handler.handleMessageRecord(toHandle, state);
					}
				}
			}).start();
		}
	}

	public State getState() {
		return state;
	}

	@Override
	public String toString() {
		int pSize = 0;
		try {
			pSize = packet.getSize();
		} catch (NotImplementedException e) {
			pSize = -1;
		}
		return String.format("MessageRecord: %d [%d >> %d] Size:%d State:%s", enqueueTime, srcAdr & 0xFFFF, dstAdr & 0xFFFF, pSize, state.name());
	}

	public void addHandler(MessageRecordHandler handler) {
		handlers.add(handler);
	}

	@Override
	public int hashCode() {
		final int prime = 31;
		int result = 1;
		result = prime * result + dstAdr;
		result = prime * result + ((packet == null) ? 0 : packet.hashCode());
		result = prime * result + srcAdr;
		return result;
	}

	@Override
	public boolean equals(Object obj) {
		if (this == obj)
			return true;
		if (obj == null)
			return false;
		if (getClass() != obj.getClass())
			return false;
		MessageRecord other = (MessageRecord) obj;
		if (dstAdr != other.dstAdr)
			return false;
		if (packet == null) {
			if (other.packet != null)
				return false;
		} else if (!packet.equals(other.packet))
			return false;
		if (srcAdr != other.srcAdr)
			return false;
		return true;
	}
	
	

}
